#!/usr/bin/env python
#
# Copyright 2008, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

r"""Tests the text output of Google C++ Testing and Mocking Framework.

To update the golden file:
googletest_output_test.py --build_dir=BUILD/DIR --gengolden
where BUILD/DIR contains the built googletest-output-test_ file.
googletest_output_test.py --gengolden
googletest_output_test.py
"""

import difflib
import os
import re
import sys
from googletest.test import gtest_test_utils


# The flag for generating the golden file
GENGOLDEN_FLAG = '--gengolden'
CATCH_EXCEPTIONS_ENV_VAR_NAME = 'GTEST_CATCH_EXCEPTIONS'

# The flag indicating stacktraces are not supported
NO_STACKTRACE_SUPPORT_FLAG = '--no_stacktrace_support'

IS_LINUX = os.name == 'posix' and os.uname()[0] == 'Linux'
IS_WINDOWS = os.name == 'nt'

GOLDEN_NAME = 'googletest-output-test-golden-lin.txt'

PROGRAM_PATH = gtest_test_utils.GetTestExecutablePath('googletest-output-test_')

# At least one command we exercise must not have the
# 'internal_skip_environment_and_ad_hoc_tests' argument.
COMMAND_LIST_TESTS = ({}, [PROGRAM_PATH, '--gtest_list_tests'])
COMMAND_WITH_COLOR = ({}, [PROGRAM_PATH, '--gtest_color=yes'])
COMMAND_WITH_TIME = ({}, [PROGRAM_PATH,
                          '--gtest_print_time',
                          'internal_skip_environment_and_ad_hoc_tests',
                          '--gtest_filter=FatalFailureTest.*:LoggingTest.*'])
COMMAND_WITH_DISABLED = (
    {}, [PROGRAM_PATH,
         '--gtest_also_run_disabled_tests',
         'internal_skip_environment_and_ad_hoc_tests',
         '--gtest_filter=*DISABLED_*'])
COMMAND_WITH_SHARDING = (
    {'GTEST_SHARD_INDEX': '1', 'GTEST_TOTAL_SHARDS': '2'},
    [PROGRAM_PATH,
     'internal_skip_environment_and_ad_hoc_tests',
     '--gtest_filter=PassingTest.*'])

GOLDEN_PATH = os.path.join(gtest_test_utils.GetSourceDir(), GOLDEN_NAME)


def ToUnixLineEnding(s):
  """Changes all Windows/Mac line endings in s to UNIX line endings."""

  return s.replace('\r\n', '\n').replace('\r', '\n')


def RemoveLocations(test_output):
  """Removes all file location info from a Google Test program's output.

  Args:
       test_output:  the output of a Google Test program.

  Returns:
       output with all file location info (in the form of
       'DIRECTORY/FILE_NAME:LINE_NUMBER: 'or
       'DIRECTORY\\FILE_NAME(LINE_NUMBER): ') replaced by
       'FILE_NAME:#: '.
  """

  return re.sub(r'.*[/\\]((googletest-output-test_|gtest).cc)(\:\d+|\(\d+\))\: ',
                r'\1:#: ', test_output)


def RemoveStackTraceDetails(output):
  """Removes all stack traces from a Google Test program's output."""

  # *? means "find the shortest string that matches".
  return re.sub(r'Stack trace:(.|\n)*?\n\n',
                'Stack trace: (omitted)\n\n', output)


def RemoveStackTraces(output):
  """Removes all traces of stack traces from a Google Test program's output."""

  # *? means "find the shortest string that matches".
  return re.sub(r'Stack trace:(.|\n)*?\n\n', '', output)


def RemoveTime(output):
  """Removes all time information from a Google Test program's output."""

  return re.sub(r'\(\d+ ms', '(? ms', output)


def RemoveTypeInfoDetails(test_output):
  """Removes compiler-specific type info from Google Test program's output.

  Args:
       test_output:  the output of a Google Test program.

  Returns:
       output with type information normalized to canonical form.
  """

  # some compilers output the name of type 'unsigned int' as 'unsigned'
  return re.sub(r'unsigned int', 'unsigned', test_output)


def NormalizeToCurrentPlatform(test_output):
  """Normalizes platform specific output details for easier comparison."""

  if IS_WINDOWS:
    # Removes the color information that is not present on Windows.
    test_output = re.sub('\x1b\\[(0;3\d)?m', '', test_output)
    # Changes failure message headers into the Windows format.
    test_output = re.sub(r': Failure\n', r': error: ', test_output)
    # Changes file(line_number) to file:line_number.
    test_output = re.sub(r'((\w|\.)+)\((\d+)\):', r'\1:\3:', test_output)

  return test_output


def RemoveTestCounts(output):
  """Removes test counts from a Google Test program's output."""

  output = re.sub(r'\d+ tests?, listed below',
                  '? tests, listed below', output)
  output = re.sub(r'\d+ FAILED TESTS',
                  '? FAILED TESTS', output)
  output = re.sub(r'\d+ tests? from \d+ test cases?',
                  '? tests from ? test cases', output)
  output = re.sub(r'\d+ tests? from ([a-zA-Z_])',
                  r'? tests from \1', output)
  return re.sub(r'\d+ tests?\.', '? tests.', output)


def RemoveMatchingTests(test_output, pattern):
  """Removes output of specified tests from a Google Test program's output.

  This function strips not only the beginning and the end of a test but also
  all output in between.

  Args:
    test_output:       A string containing the test output.
    pattern:           A regex string that matches names of test cases or
                       tests to remove.

  Returns:
    Contents of test_output with tests whose names match pattern removed.
  """

  test_output = re.sub(
      r'.*\[ RUN      \] .*%s(.|\n)*?\[(  FAILED  |       OK )\] .*%s.*\n' % (
          pattern, pattern),
      '',
      test_output)
  return re.sub(r'.*%s.*\n' % pattern, '', test_output)


def NormalizeOutput(output):
  """Normalizes output (the output of googletest-output-test_.exe)."""

  output = ToUnixLineEnding(output)
  output = RemoveLocations(output)
  output = RemoveStackTraceDetails(output)
  output = RemoveTime(output)
  return output


def GetShellCommandOutput(env_cmd):
  """Runs a command in a sub-process, and returns its output in a string.

  Args:
    env_cmd: The shell command. A 2-tuple where element 0 is a dict of extra
             environment variables to set, and element 1 is a string with
             the command and any flags.

  Returns:
    A string with the command's combined standard and diagnostic output.
  """

  # Spawns cmd in a sub-process, and gets its standard I/O file objects.
  # Set and save the environment properly.
  environ = os.environ.copy()
  environ.update(env_cmd[0])
  p = gtest_test_utils.Subprocess(env_cmd[1], env=environ)

  return p.output


def GetCommandOutput(env_cmd):
  """Runs a command and returns its output with all file location
  info stripped off.

  Args:
    env_cmd:  The shell command. A 2-tuple where element 0 is a dict of extra
              environment variables to set, and element 1 is a string with
              the command and any flags.
  """

  # Disables exception pop-ups on Windows.
  environ, cmdline = env_cmd
  environ = dict(environ)  # Ensures we are modifying a copy.
  environ[CATCH_EXCEPTIONS_ENV_VAR_NAME] = '1'
  return NormalizeOutput(GetShellCommandOutput((environ, cmdline)))


def GetOutputOfAllCommands():
  """Returns concatenated output from several representative commands."""

  return (GetCommandOutput(COMMAND_WITH_COLOR) +
          GetCommandOutput(COMMAND_WITH_TIME) +
          GetCommandOutput(COMMAND_WITH_DISABLED) +
          GetCommandOutput(COMMAND_WITH_SHARDING))


test_list = GetShellCommandOutput(COMMAND_LIST_TESTS)
SUPPORTS_DEATH_TESTS = 'DeathTest' in test_list
SUPPORTS_TYPED_TESTS = 'TypedTest' in test_list
SUPPORTS_THREADS = 'ExpectFailureWithThreadsTest' in test_list
SUPPORTS_STACK_TRACES = NO_STACKTRACE_SUPPORT_FLAG not in sys.argv

CAN_GENERATE_GOLDEN_FILE = (SUPPORTS_DEATH_TESTS and
                            SUPPORTS_TYPED_TESTS and
                            SUPPORTS_THREADS and
                            SUPPORTS_STACK_TRACES)

class GTestOutputTest(gtest_test_utils.TestCase):
  def RemoveUnsupportedTests(self, test_output):
    if not SUPPORTS_DEATH_TESTS:
      test_output = RemoveMatchingTests(test_output, 'DeathTest')
    if not SUPPORTS_TYPED_TESTS:
      test_output = RemoveMatchingTests(test_output, 'TypedTest')
      test_output = RemoveMatchingTests(test_output, 'TypedDeathTest')
      test_output = RemoveMatchingTests(test_output, 'TypeParamDeathTest')
    if not SUPPORTS_THREADS:
      test_output = RemoveMatchingTests(test_output,
                                        'ExpectFailureWithThreadsTest')
      test_output = RemoveMatchingTests(test_output,
                                        'ScopedFakeTestPartResultReporterTest')
      test_output = RemoveMatchingTests(test_output,
                                        'WorksConcurrently')
    if not SUPPORTS_STACK_TRACES:
      test_output = RemoveStackTraces(test_output)

    return test_output

  def testOutput(self):
    output = GetOutputOfAllCommands()

    golden_file = open(GOLDEN_PATH, 'rb')
    # A mis-configured source control system can cause \r appear in EOL
    # sequences when we read the golden file irrespective of an operating
    # system used. Therefore, we need to strip those \r's from newlines
    # unconditionally.
    golden = ToUnixLineEnding(golden_file.read().decode())
    golden_file.close()

    # We want the test to pass regardless of certain features being
    # supported or not.

    # We still have to remove type name specifics in all cases.
    normalized_actual = RemoveTypeInfoDetails(output)
    normalized_golden = RemoveTypeInfoDetails(golden)

    if CAN_GENERATE_GOLDEN_FILE:
      self.assertEqual(normalized_golden, normalized_actual,
                       '\n'.join(difflib.unified_diff(
                           normalized_golden.split('\n'),
                           normalized_actual.split('\n'),
                           'golden', 'actual')))
    else:
      normalized_actual = NormalizeToCurrentPlatform(
          RemoveTestCounts(normalized_actual))
      normalized_golden = NormalizeToCurrentPlatform(
          RemoveTestCounts(self.RemoveUnsupportedTests(normalized_golden)))

      # This code is very handy when debugging golden file differences:
      if os.getenv('DEBUG_GTEST_OUTPUT_TEST'):
        open(os.path.join(
            gtest_test_utils.GetSourceDir(),
            '_googletest-output-test_normalized_actual.txt'), 'wb').write(
                normalized_actual)
        open(os.path.join(
            gtest_test_utils.GetSourceDir(),
            '_googletest-output-test_normalized_golden.txt'), 'wb').write(
                normalized_golden)

      self.assertEqual(normalized_golden, normalized_actual)


if __name__ == '__main__':
  if NO_STACKTRACE_SUPPORT_FLAG in sys.argv:
    # unittest.main() can't handle unknown flags
    sys.argv.remove(NO_STACKTRACE_SUPPORT_FLAG)

  if GENGOLDEN_FLAG in sys.argv:
    if CAN_GENERATE_GOLDEN_FILE:
      output = GetOutputOfAllCommands()
      golden_file = open(GOLDEN_PATH, 'wb')
      golden_file.write(output.encode())
      golden_file.close()
    else:
      message = (
          """Unable to write a golden file when compiled in an environment
that does not support all the required features (death tests,
typed tests, stack traces, and multiple threads).
Please build this test and generate the golden file using Blaze on Linux.""")

      sys.stderr.write(message)
      sys.exit(1)
  else:
    gtest_test_utils.Main()



#!/usr/bin/env python3
"""
Setup script for iowarp-core package.
Builds and installs C++ components using CMake in the correct order.
"""

import os
import sys
import subprocess
import shutil
import platform
from pathlib import Path
from setuptools import setup, Extension
from setuptools.dist import Distribution
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel


class CMakeExtension(Extension):
    """Extension class for CMake-based C++ projects."""

    def __init__(self, name, sourcedir="", repo_url="", **kwargs):
        super().__init__(name, sources=[], **kwargs)
        self.sourcedir = os.path.abspath(sourcedir)
        self.repo_url = repo_url


class CustomSDist(sdist):
    """Custom sdist command that ensures git submodules are fully included."""

    def run(self):
        """Ensure git submodules are initialized before creating source distribution."""
        # Check if we're in a git repository
        if os.path.exists(".git"):
            print("\n" + "="*60)
            print("Ensuring git submodules are included in source distribution")
            print("="*60 + "\n")

            try:
                # Initialize and update submodules to ensure they're present
                subprocess.check_call(
                    ["git", "submodule", "update", "--init", "--recursive"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT
                )
                print("✓ Git submodules initialized successfully")

                # List submodules to verify
                result = subprocess.run(
                    ["git", "submodule", "status"],
                    capture_output=True,
                    text=True,
                    check=True
                )
                if result.stdout:
                    print("\nSubmodules found:")
                    for line in result.stdout.strip().split('\n'):
                        print(f"  {line}")

                print("\nNote: MANIFEST.in will include submodule files in the tarball")
                print("")

            except subprocess.CalledProcessError as e:
                print(f"Warning: Could not initialize git submodules: {e}")
                print("Source distribution may be incomplete!")
        else:
            print("Not a git repository - skipping submodule initialization")

        # Call parent sdist command to create the distribution
        super().run()


class CMakeBuild(build_ext):
    """Custom build command that builds IOWarp core using CMake presets."""

    # Single repository for all components
    REPO_URL = "https://github.com/iowarp/core"

    def run(self):
        """Build IOWarp core following the quick installation steps."""
        try:
            subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build iowarp-core. "
                "Install with: pip install cmake"
            )

        # Create build directory
        build_temp = Path(self.build_temp).absolute()
        build_temp.mkdir(parents=True, exist_ok=True)

        # Build the unified core
        self.build_iowarp_core(build_temp)

        # If bundling binaries, copy them to the package directory
        if os.environ.get("IOWARP_BUNDLE_BINARIES", "OFF").upper() == "ON":
            self.copy_binaries_to_package(build_temp)

    def build_iowarp_core(self, build_temp):
        """Build IOWarp core using install.sh script."""
        print(f"\n{'='*60}")
        print(f"Building IOWarp Core using install.sh")
        print(f"{'='*60}\n")

        # Determine install prefix based on whether we're bundling binaries
        # For IOWarp Core, install directly to the Python environment (sys.prefix)
        # This is the standard approach for packages with C++ libraries that need to be
        # found by CMake and linked by other applications.
        # Libraries → {sys.prefix}/lib/
        # Headers → {sys.prefix}/include/
        # CMake configs → {sys.prefix}/lib/cmake/
        bundle_binaries = os.environ.get("IOWARP_BUNDLE_BINARIES", "OFF").upper() == "ON"
        if bundle_binaries:
            # Install to a staging directory that we'll copy into the wheel
            # (Not recommended for IOWarp - use IOWARP_BUNDLE_BINARIES=ON to enable)
            install_prefix = build_temp / "install"
        else:
            # Install directly to Python environment prefix
            # This ensures libraries/headers are in standard locations
            install_prefix = Path(sys.prefix).absolute()

        print(f"Install prefix: {install_prefix}")

        # Find install.sh in package root
        package_root = Path(__file__).parent.absolute()
        install_script = package_root / "install.sh"

        if not install_script.exists():
            raise RuntimeError(f"install.sh not found at {install_script}")

        # Make install.sh executable
        install_script.chmod(0o755)

        # Prepare environment for install.sh
        env = os.environ.copy()
        env["INSTALL_PREFIX"] = str(install_prefix)

        # Determine number of parallel jobs
        if hasattr(self, "parallel") and self.parallel:
            env["BUILD_JOBS"] = str(self.parallel)
        else:
            import multiprocessing
            env["BUILD_JOBS"] = str(multiprocessing.cpu_count())

        print(f"\nRunning install.sh with:")
        print(f"  INSTALL_PREFIX={env['INSTALL_PREFIX']}")
        print(f"  BUILD_JOBS={env['BUILD_JOBS']}\n")

        # Run install.sh and capture output for debugging
        result = subprocess.run(
            [str(install_script)],
            cwd=package_root,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True
        )

        # Print all install.sh output
        if result.stdout:
            print(result.stdout)

        # Check for errors
        if result.returncode != 0:
            print(f"\nERROR: install.sh failed with exit code {result.returncode}\n")
            sys.exit(result.returncode)

        print(f"\nIOWarp core built and installed successfully!\n")

    def copy_binaries_to_package(self, build_temp):
        """Copy built binaries and headers into the Python package for wheel bundling."""
        print("\n" + "="*60)
        print("Copying binaries to package directory")
        print("="*60 + "\n")

        install_prefix = build_temp / "install"
        package_dir = Path(self.build_lib) / "iowarp_core"

        # Create directories in the package
        lib_dir = package_dir / "lib"
        include_dir = package_dir / "include"
        bin_dir = package_dir / "bin"

        lib_dir.mkdir(parents=True, exist_ok=True)
        include_dir.mkdir(parents=True, exist_ok=True)
        bin_dir.mkdir(parents=True, exist_ok=True)

        # Copy libraries
        src_lib_dir = install_prefix / "lib"
        if src_lib_dir.exists():
            print(f"Copying libraries from {src_lib_dir} to {lib_dir}")
            for lib_file in src_lib_dir.rglob("*"):
                if lib_file.is_file() or lib_file.is_symlink():
                    # Copy .so, .a, and .dylib files
                    if lib_file.suffix in [".so", ".a", ".dylib"] or ".so." in lib_file.name:
                        rel_path = lib_file.relative_to(src_lib_dir)
                        dest = lib_dir / rel_path
                        dest.parent.mkdir(parents=True, exist_ok=True)
                        # Remove existing file/symlink to avoid conflicts
                        if dest.exists() or dest.is_symlink():
                            dest.unlink()
                        # Copy file or symlink
                        if lib_file.is_symlink():
                            os.symlink(os.readlink(lib_file), dest)
                        else:
                            shutil.copy2(lib_file, dest)
                        print(f"  Copied: {rel_path}")

        # Copy lib64 if it exists (some systems use lib64)
        src_lib64_dir = install_prefix / "lib64"
        if src_lib64_dir.exists():
            print(f"Copying libraries from {src_lib64_dir} to {lib_dir}")
            for lib_file in src_lib64_dir.rglob("*"):
                if lib_file.is_file() or lib_file.is_symlink():
                    if lib_file.suffix in [".so", ".a", ".dylib"] or ".so." in lib_file.name:
                        rel_path = lib_file.relative_to(src_lib64_dir)
                        dest = lib_dir / rel_path
                        dest.parent.mkdir(parents=True, exist_ok=True)
                        # Remove existing file/symlink to avoid conflicts
                        if dest.exists() or dest.is_symlink():
                            dest.unlink()
                        # Copy file or symlink
                        if lib_file.is_symlink():
                            os.symlink(os.readlink(lib_file), dest)
                        else:
                            shutil.copy2(lib_file, dest)
                        print(f"  Copied: {rel_path}")

        # Copy headers
        src_include_dir = install_prefix / "include"
        if src_include_dir.exists():
            print(f"Copying headers from {src_include_dir} to {include_dir}")
            for header_file in src_include_dir.rglob("*"):
                if header_file.is_file():
                    rel_path = header_file.relative_to(src_include_dir)
                    dest = include_dir / rel_path
                    dest.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy2(header_file, dest)

        # Copy binaries/executables
        src_bin_dir = install_prefix / "bin"
        if src_bin_dir.exists():
            print(f"Copying binaries from {src_bin_dir} to {bin_dir}")
            # Binaries to exclude from distribution (test executables)
            exclude_binaries = {
                "test_binary_assim",
                "test_error_handling",
                "test_hdf5_assim",
                "test_range_assim",
            }
            for bin_file in src_bin_dir.rglob("*"):
                if bin_file.is_file():
                    # Skip test binaries
                    if bin_file.name in exclude_binaries:
                        print(f"  Skipped test binary: {bin_file.name}")
                        continue

                    rel_path = bin_file.relative_to(src_bin_dir)
                    dest = bin_dir / rel_path
                    dest.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy2(bin_file, dest)
                    # Make executable
                    dest.chmod(dest.stat().st_mode | 0o111)
                    print(f"  Copied: {rel_path}")

        # Copy conda dependencies
        conda_prefix = os.environ.get("CONDA_PREFIX")
        if conda_prefix:
            print(f"\n" + "="*60)
            print("Copying conda dependencies")
            print("="*60 + "\n")

            conda_lib_dir = Path(conda_prefix) / "lib"
            if conda_lib_dir.exists():
                # List of library patterns to copy (dependencies needed by IOWarp)
                lib_patterns = [
                    "libboost_*.so*",
                    "libhdf5*.so*",
                    "libmpi*.so*",
                    "libzmq*.so*",
                    "libsodium.so*",  # Required by ZeroMQ
                    "libyaml*.so*",  # Note: libyaml-cpp is built from source (see skip_libs below)
                    "libz.so*",
                    "libsz.so*",
                    "libaec.so*",
                    "libcurl.so*",
                    "libssh2.so*",  # Required by libcurl
                    "libnghttp2.so*",  # Required by libcurl
                    "libssl.so*",  # OpenSSL SSL library
                    "libcrypto.so*",  # OpenSSL crypto library
                    "libopen-*.so*",  # OpenMPI libraries
                    "libpmix*.so*",  # PMIx for OpenMPI
                    "libhwloc*.so*",  # Hardware locality for MPI
                    "libevent*.so*",  # Event notification library
                    "libfabric*.so*",  # Networking for MPI
                    "libefa.so*",  # Elastic Fabric Adapter (if available)
                    "libpsm*.so*",  # Intel PSM/PSM2 (if available)
                    "libucx*.so*",  # Unified Communication X
                    "libucp*.so*",  # UCX Protocol layer
                    "libucc*.so*",  # Unified Collective Communication
                    "libucs*.so*",  # UCX Services layer
                    "libuct*.so*",  # UCX Transport layer
                    "libucm*.so*",  # UCX Memory layer
                    "libicu*.so*",  # ICU (International Components for Unicode)
                    "libnuma*.so*",  # NUMA support
                    "librdmacm.so*",  # RDMA connection manager
                    "libibverbs.so*",  # InfiniBand verbs
                    "libstdc++.so*",
                    "libgcc_s.so*",
                    "libgfortran.so*",
                    "libquadmath.so*",
                ]

                copied_libs = set()
                # Libraries we build from source and should not copy from conda
                skip_libs = {
                    "libyaml-cpp.so",
                    "libyaml-cpp.so.0.8",
                    "libyaml-cpp.so.0.8.0",
                }

                for pattern in lib_patterns:
                    for lib_file in conda_lib_dir.glob(pattern):
                        lib_name = lib_file.name

                        # Skip libraries we build from source
                        if lib_name in skip_libs:
                            print(f"  Skipping conda {lib_name} (using built-from-source version)")
                            continue

                        if lib_file.is_file() and not lib_file.is_symlink():
                            if lib_name not in copied_libs:
                                dest = lib_dir / lib_name
                                shutil.copy2(lib_file, dest)
                                copied_libs.add(lib_name)
                                print(f"  Copied conda dependency: {lib_name}")
                        elif lib_file.is_symlink():
                            # Copy symlinks as well
                            if lib_name not in copied_libs:
                                dest = lib_dir / lib_name
                                # Remove existing file/symlink to avoid conflicts
                                if dest.exists() or dest.is_symlink():
                                    dest.unlink()
                                target = lib_file.readlink()
                                # If target is relative, keep it relative
                                if not target.is_absolute():
                                    dest.symlink_to(target)
                                else:
                                    # If absolute, just copy the file it points to
                                    shutil.copy2(lib_file, dest)
                                copied_libs.add(lib_name)
                                print(f"  Copied conda dependency (symlink): {lib_name}")

                print(f"\nTotal conda dependencies copied: {len(copied_libs)}")

        # Fix RPATH in all bundled libraries to prefer bundled dependencies
        print(f"\n" + "="*60)
        print("Fixing RPATH in bundled libraries")
        print("="*60 + "\n")

        # Check if patchelf is available
        patchelf_available = True
        try:
            subprocess.run(["patchelf", "--version"],
                         capture_output=True, check=True)
        except (subprocess.CalledProcessError, FileNotFoundError):
            print("  Warning: patchelf not found, skipping RPATH fixes")
            print("  Libraries may fail to find bundled dependencies")
            patchelf_available = False

        if patchelf_available:
            # Fix RPATH for all .so files in lib directory
            for lib_file in lib_dir.rglob("*.so*"):
                if lib_file.is_file() and not lib_file.is_symlink():
                    try:
                        # Set RPATH to look in same directory first
                        # $ORIGIN means the directory containing the library
                        subprocess.run([
                            "patchelf",
                            "--set-rpath", "$ORIGIN:$ORIGIN/..:$ORIGIN/../lib",
                            "--force-rpath",
                            str(lib_file)
                        ], capture_output=True, check=True)
                        print(f"  Fixed RPATH: {lib_file.name}")
                    except subprocess.CalledProcessError as e:
                        # Some files may not be ELF files or may not have RPATH
                        # This is fine, just skip them
                        pass

            # Fix RPATH for binaries
            for bin_file in bin_dir.rglob("*"):
                if bin_file.is_file() and not bin_file.is_symlink():
                    try:
                        # Set RPATH to look in ../lib
                        subprocess.run([
                            "patchelf",
                            "--set-rpath", "$ORIGIN/../lib",
                            "--force-rpath",
                            str(bin_file)
                        ], capture_output=True, check=True)
                        print(f"  Fixed RPATH: bin/{bin_file.name}")
                    except subprocess.CalledProcessError:
                        pass

        print("\nBinary copying complete!\n")


class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform-specific tags."""

    def has_ext_modules(self):
        """Always return True to indicate this is a platform-specific package.

        This forces setuptools to generate platform-specific wheel tags instead of
        pure-python 'any' tags. Required for packages with compiled C/C++ extensions.
        """
        return True


class CustomBdistWheel(_bdist_wheel):
    """Custom bdist_wheel that automatically generates manylinux tags.

    This eliminates the need for post-build wheel renaming scripts and ensures
    proper platform tags are set during the build process.
    """

    def finalize_options(self):
        """Override to set platform-specific wheel tags based on the target system."""
        super().finalize_options()

        # Only apply manylinux tags on Linux
        if platform.system() == 'Linux':
            machine = platform.machine()

            # Map machine architecture to manylinux platform tag
            # manylinux_2_17 is compatible with most modern Linux systems (RHEL 7+, Ubuntu 16.04+)
            if machine == 'x86_64':
                self.plat_name = 'manylinux_2_17_x86_64'
            elif machine == 'aarch64':
                self.plat_name = 'manylinux_2_17_aarch64'
            elif machine == 'ppc64le':
                self.plat_name = 'manylinux_2_17_ppc64le'
            elif machine == 's390x':
                self.plat_name = 'manylinux_2_17_s390x'
            else:
                # For unknown architectures, use the generic linux tag
                print(f"Warning: Unknown architecture {machine}, using generic linux tag")
                self.plat_name = f'linux_{machine}'


# Create extensions list
# Always include the CMake build extension so that source distributions work correctly.
# The IOWARP_BUNDLE_BINARIES flag controls whether binaries are bundled into the wheel
# or installed to the system prefix (for source installs).
ext_modules = [
    CMakeExtension(
        "iowarp_core._native",
        sourcedir=".",
    )
]
cmdclass = {
    "build_ext": CMakeBuild,
    "sdist": CustomSDist,
    "bdist_wheel": CustomBdistWheel,
}


if __name__ == "__main__":
    setup(
        ext_modules=ext_modules,
        cmdclass=cmdclass,
        distclass=BinaryDistribution,
    )



from code_generators.c_macro.macro_generator import CppMacroGenerator
import pathlib, os

PROJECT_ROOT=pathlib.Path(__file__).parent.parent.resolve()

DATA_STRUCTURE_TEMPLATES='include/hermes_shm/data_structures/internal/template'
DATA_STRUCTURE_INTERNAL='include/hermes_shm/data_structures/internal'

CppMacroGenerator().generate(
    os.path.join(PROJECT_ROOT,
                 f"{DATA_STRUCTURE_TEMPLATES}/shm_container_base_template.h"),
    os.path.join(PROJECT_ROOT,
                f"{DATA_STRUCTURE_INTERNAL}/shm_container_macro.h"),
    "HIPC_CONTAINER_TEMPLATE",
    ["CLASS_NAME", "TYPED_CLASS"],
    ["TYPE_UNWRAP", "TYPE_UNWRAP"],
    "HSHM_DATA_STRUCTURES_INTERNAL_SHM_CONTAINER_MACRO_H_")



##############################################################################
# Copyright by The HDF Group.                                                #
#              Helmholtz-Zentrum Dresden - Rossendorf,                       #
#                Computational Radiation Physics                             #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
from __future__ import absolute_import, division, print_function

from hdf_compass.compass_model.test import container, store
from hdf_compass.adios_model import ADIOSGroup, ADIOSStore
from hdf_compass.utils import data_url

import os

url = os.path.join(data_url(), "adios", "adios_test.bp")

s = store(ADIOSStore, url)
c = container(ADIOSStore, url, ADIOSGroup, "")




from jarvis_cd.core.pkg import Application
from jarvis_cd.shell import Exec, MpiExecInfo
import os

class WrpCteBench(Application):
    """
    CTE Core Benchmark Application

    This application runs benchmarks for Put, Get, and GetTagSize operations
    in the Content Transfer Engine (CTE) with MPI support for parallel I/O.

    The benchmark measures:
    - Put operations: Write data to CTE
    - Get operations: Read data from CTE
    - PutGet operations: Combined write and read operations

    It supports async operation depth configuration for testing concurrent I/O.
    """

    def _init(self):
        """
        Initialize the WrpCteBench application.

        This method is called during application initialization.
        """
        self.benchmark_executable = 'wrp_cte_bench'
        self.output_file = None

    def _configure_menu(self):
        """
        Configure the application menu.

        Returns:
            List[Dict]: Configuration menu options for the benchmark.
        """
        return [
            {
                'name': 'test_case',
                'msg': 'Benchmark test case to run',
                'type': str,
                'choices': ['Put', 'Get', 'PutGet'],
                'default': 'Put',
                'help': 'Put: Write benchmark, Get: Read benchmark, PutGet: Combined write+read benchmark'
            },
            {
                'name': 'depth',
                'msg': 'Number of async requests to generate',
                'type': int,
                'default': 4,
                'help': 'Number of concurrent async I/O operations (e.g., 4 means 4 operations in flight)'
            },
            {
                'name': 'io_size',
                'msg': 'Size of I/O operations',
                'type': str,
                'default': '1m',
                'help': 'I/O size with suffix: k/K (KB), m/M (MB), g/G (GB). Examples: 4k, 1m, 2g'
            },
            {
                'name': 'io_count',
                'msg': 'Number of I/O operations to generate per node',
                'type': int,
                'default': 100,
                'help': 'Total number of I/O operations each MPI rank will perform'
            },
            {
                'name': 'nprocs',
                'msg': 'Number of MPI processes',
                'type': int,
                'default': 1,
                'help': 'Number of MPI processes to use for parallel I/O'
            },
            {
                'name': 'ppn',
                'msg': 'Processes per node',
                'type': int,
                'default': 1,
                'help': 'Number of MPI processes per node'
            },
            {
                'name': 'output_file',
                'msg': 'Output file for benchmark results',
                'type': str,
                'default': '',
                'help': 'Path to save benchmark results. If empty, results are printed to stdout'
            },
            {
                'name': 'init_runtime',
                'msg': 'Initialize Chimaera runtime (otherwise assumes runtime already running)',
                'type': bool,
                'default': False,
                'help': 'Set to True to initialize runtime, False to only initialize client'
            }
        ]

    def _configure(self, **kwargs):
        """
        Configure the CTE benchmark application with provided keyword arguments.

        This method sets up the benchmark executable path and output file.

        Args:
            **kwargs: Configuration arguments from _configure_menu.
        """
        self.log("Configuring CTE benchmark application...")

        # Validate test_case
        if self.config['test_case'] not in ['Put', 'Get', 'PutGet']:
            raise ValueError(f"Invalid test_case: {self.config['test_case']}. Must be Put, Get, or PutGet")

        # Validate depth
        if self.config['depth'] <= 0:
            raise ValueError(f"Invalid depth: {self.config['depth']}. Must be > 0")

        # Validate io_count
        if self.config['io_count'] <= 0:
            raise ValueError(f"Invalid io_count: {self.config['io_count']}. Must be > 0")

        # Validate io_size format (should have k/K, m/M, g/G suffix or be a number)
        io_size_str = str(self.config['io_size']).lower()
        if not (io_size_str[-1] in ['k', 'm', 'g'] or io_size_str.isdigit()):
            self.log(f"Warning: io_size '{self.config['io_size']}' should end with k/K, m/M, or g/G suffix")

        # Set output file if specified
        if self.config['output_file']:
            self.output_file = os.path.join(self.shared_dir, self.config['output_file'])
            self.log(f"Benchmark results will be saved to: {self.output_file}")
        else:
            self.output_file = None
            self.log("Benchmark results will be printed to stdout")

        # Set environment variables for the benchmark
        self.setenv('CTE_BENCH_TEST_CASE', self.config['test_case'])
        self.setenv('CTE_BENCH_DEPTH', str(self.config['depth']))
        self.setenv('CTE_BENCH_IO_SIZE', str(self.config['io_size']))
        self.setenv('CTE_BENCH_IO_COUNT', str(self.config['io_count']))

        # Set CHIMAERA_WITH_RUNTIME environment variable based on configuration
        if self.config['init_runtime']:
            self.setenv('CHIMAERA_WITH_RUNTIME', '1')
            self.log("Runtime initialization enabled (CHIMAERA_WITH_RUNTIME=1)")
        else:
            self.setenv('CHIMAERA_WITH_RUNTIME', '0')
            self.log("Runtime initialization disabled (CHIMAERA_WITH_RUNTIME=0)")

        self.log("CTE benchmark configuration completed successfully")

    def start(self):
        """
        Run the CTE benchmark application.

        This method executes the benchmark with MPI support and configured parameters.
        """
        self.log(f"Starting CTE benchmark: {self.config['test_case']}")

        # Build command line arguments
        cmd = [
            self.benchmark_executable,
            self.config['test_case'],
            str(self.config['depth']),
            str(self.config['io_size']),
            str(self.config['io_count'])
        ]

        # Execute with MPI if nprocs > 1
        if self.config['nprocs'] > 1:
            self.log(f"Running benchmark with MPI: {self.config['nprocs']} processes, {self.config['ppn']} per node")

            exec_info = MpiExecInfo(
                env=self.mod_env,
                hostfile=self.hostfile,
                nprocs=self.config['nprocs'],
                ppn=self.config['ppn']
            )
        else:
            self.log("Running benchmark without MPI (single process)")
            from jarvis_cd.shell import LocalExecInfo
            exec_info = LocalExecInfo(env=self.mod_env)

        # Execute the benchmark
        cmd_str = ' '.join(cmd)

        if self.output_file:
            # Redirect output to file
            cmd_str += f' > {self.output_file} 2>&1'
            self.log(f"Executing: {cmd_str}")
            Exec(cmd_str, exec_info).run()
            self.log(f"Benchmark completed. Results saved to: {self.output_file}")
        else:
            # Print to stdout
            self.log(f"Executing: {cmd_str}")
            Exec(cmd_str, exec_info).run()
            self.log("Benchmark completed")

    def stop(self):
        """
        Stop the benchmark application.

        Since this is an application that runs to completion, this method
        is typically not needed but is provided for consistency.
        """
        self.log("WrpCteBench is an application - it runs to completion")
        return True

    def clean(self):
        """
        Clean up benchmark output files.
        """
        self.log("Cleaning up CTE benchmark output files...")

        # Clean output file if it exists
        if self.output_file and os.path.exists(self.output_file):
            try:
                os.remove(self.output_file)
                self.log(f"Removed benchmark output file: {self.output_file}")
            except Exception as e:
                self.log(f"Error removing output file: {e}")

        self.log("CTE benchmark cleanup completed")



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
"""
Implementation of utils and helper functions
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import os

import logging
log = logging.getLogger(__name__)

is_darwin = sys.platform == 'darwin'
is_win = sys.platform == 'win32'
is_linux = sys.platform == 'linux2'


def url2path(url):
    """ Helper function that returns the file path from an url, dealing with Windows peculiarities """
    if is_win:
        # Remove 'file:///' prefix and handle Windows paths
        path = url.replace('file://', '')
        # Convert forward slashes to backslashes for Windows
        path = path.replace('/', '\\')
        return path
    else:
        return url.replace('file://', '')


def path2url(path):
    """ Helper function that returns the url from a file path, dealing with Windows peculiarities """
    if is_win:
        # Convert backslashes to forward slashes for URLs
        path = path.replace('\\', '/')
        return 'file:///' + path
    else:
        return 'file://' + path


def data_url():
    """ Helper function used to return the url to the project data folder """
    prj_root_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, os.pardir))
    data_folder = os.path.join(prj_root_folder, "data")

    if not os.path.exists(data_folder):
        raise RuntimeError("data path %s does not exist" % data_folder)

    return path2url(data_folder)



#!/usr/bin/env python3
"""
Test suite for IOWarp CTE MCP Server.

This test file mimics MCP behavior by using the MCP client library to call
all tools and capture their inputs and outputs. This allows verification
that all Context Interface operations work correctly.
"""

import asyncio
import json
import sys
from pathlib import Path
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client

# Colors for output
class Colors:
    PASS = '\033[92m'  # Green
    FAIL = '\033[91m'  # Red
    WARN = '\033[93m'  # Yellow
    INFO = '\033[94m'  # Blue
    RESET = '\033[0m'   # Reset

def print_test(test_name, status, details=""):
    """Print test result with colors."""
    if status == "PASS":
        symbol = "✓"
        color = Colors.PASS
    elif status == "FAIL":
        symbol = "✗"
        color = Colors.FAIL
    elif status == "EXPECTED_FAILURE":
        symbol="⚠"
        color = Colors.WARN
    else:
        symbol = "?"
        color = Colors.INFO
    
    print(f"{color}{symbol} {test_name}{Colors.RESET}")
    if details:
        print(f"    {details}")

async def test_tool(session, tool_name, arguments, expected_fields=None):
    """Test a single MCP tool and return result."""
    print(f"\n🔧 Testing: {tool_name}")
    print(f"   Input: {json.dumps(arguments, indent=2)}")
    
    try:
        result = await session.call_tool(tool_name, arguments)
        output = result.content[0].text if result.content else ""
        
        # Parse JSON output
        try:
            output_json = json.loads(output)
        except json.JSONDecodeError:
            output_json = {"raw_output": output}
        
        print(f"   Output: {json.dumps(output_json, indent=2)}")
        
        # Check if expected fields are present
        if expected_fields:
            missing = [f for f in expected_fields if f not in output_json]
            if missing:
                print_test(f"{tool_name}", "FAIL", f"Missing fields: {missing}")
                return False
        
        # Check for error in output
        if "error" in output_json or "success" in output_json and not output_json.get("success", True):
            print_test(f"{tool_name}", "EXPECTED_FAILURE", "Tool returned error (may be expected without runtime)")
            return "EXPECTED_FAILURE"
        
        print_test(f"{tool_name}", "PASS")
        return True
        
    except Exception as e:
        print(f"   Error: {type(e).__name__}: {str(e)}")
        print_test(f"{tool_name}", "FAIL", f"Exception: {str(e)}")
        return False

async def main():
    """Run all MCP tool tests."""
    print("=" * 80)
    print("IOWarp CTE MCP Server - Test Suite")
    print("=" * 80)
    
    # Setup server parameters
    server_script = Path(__file__).parent / "server.py"
    server_params = StdioServerParameters(
        command="python3",
        args=[str(server_script)],
        env={
            "PYTHONPATH": "/workspace/build/bin"
        }
    )
    
    results = {
        "PASS": [],
        "FAIL": [],
        "EXPECTED_FAILURE": []
    }
    
    async with stdio_client(server_params) as (read, write):
        async with ClientSession(read, write) as session:
            await session.initialize()
            
            # List all available tools
            tools_response = await session.list_tools()
            print(f"\n📋 Found {len(tools_response.tools)} available tools")
            
            # Test 1: Runtime Management Tools (always work)
            print("\n" + "=" * 80)
            print("1. Runtime Management Tools")
            print("=" * 80)
            
            result = await test_tool(session, "get_client_status", {}, 
                                   expected_fields=["available"])
            results["PASS" if result else "FAIL"].append("get_client_status")
            
            result = await test_tool(session, "get_cte_types", {},
                                   expected_fields=["available"])
            results["PASS" if result else "FAIL"].append("get_cte_types")
            
            # Test 2: Initialize Runtime (may succeed or fail depending on environment)
            print("\n" + "=" * 80)
            print("2. Runtime Initialization")
            print("=" * 80)
            
            result = await test_tool(session, "initialize_cte_runtime", {},
                                   expected_fields=["success", "runtime_init", "client_init", "cte_init"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("initialize_cte_runtime")
            elif result:
                results["PASS"].append("initialize_cte_runtime")
            else:
                results["FAIL"].append("initialize_cte_runtime")
            
            # Test 3: Context Interface Operations
            print("\n" + "=" * 80)
            print("3. Context Interface Operations")
            print("=" * 80)
            
            # 3.1: Put Blob (context_bundle)
            result = await test_tool(session, "put_blob", {
                "tag_name": "test_tag",
                "blob_name": "test_blob",
                "data": "Hello, World!"
            }, expected_fields=["tag_name", "blob_name", "success"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("put_blob")
            elif result:
                results["PASS"].append("put_blob")
            else:
                results["FAIL"].append("put_blob")
            
            # 3.2: List Blobs (context_query - list)
            result = await test_tool(session, "list_blobs_in_tag", {
                "tag_name": "test_tag"
            }, expected_fields=["tag_name", "blobs", "count"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("list_blobs_in_tag")
            elif result:
                results["PASS"].append("list_blobs_in_tag")
            else:
                results["FAIL"].append("list_blobs_in_tag")
            
            # 3.3: Get Blob Size (context_query - get size)
            result = await test_tool(session, "get_blob_size", {
                "tag_name": "test_tag",
                "blob_name": "test_blob"
            }, expected_fields=["tag_name", "blob_name", "size"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("get_blob_size")
            elif result:
                results["PASS"].append("get_blob_size")
            else:
                results["FAIL"].append("get_blob_size")
            
            # 3.4: Get Blob (context_query - get data)
            result = await test_tool(session, "get_blob", {
                "tag_name": "test_tag",
                "blob_name": "test_blob"
            }, expected_fields=["tag_name", "blob_name"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("get_blob")
            elif result:
                results["PASS"].append("get_blob")
            else:
                results["FAIL"].append("get_blob")
            
            # 3.5: Delete Blob (context_delete)
            result = await test_tool(session, "delete_blob", {
                "tag_name": "test_tag",
                "blob_name": "test_blob"
            }, expected_fields=["tag_name", "blob_name", "success"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("delete_blob")
            elif result:
                results["PASS"].append("delete_blob")
            else:
                results["FAIL"].append("delete_blob")
            
            # Test 4: Additional CTE Operations
            print("\n" + "=" * 80)
            print("4. Additional CTE Operations")
            print("=" * 80)
            
            result = await test_tool(session, "tag_query", {
                "tag_regex": ".*",
                "max_tags": 10
            }, expected_fields=["tag_regex", "tags", "count"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("tag_query")
            elif result:
                results["PASS"].append("tag_query")
            else:
                results["FAIL"].append("tag_query")
            
            result = await test_tool(session, "blob_query", {
                "tag_regex": ".*",
                "blob_regex": ".*",
                "max_blobs": 10
            }, expected_fields=["tag_regex", "blob_regex", "blobs", "count"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("blob_query")
            elif result:
                results["PASS"].append("blob_query")
            else:
                results["FAIL"].append("blob_query")
            
            result = await test_tool(session, "poll_telemetry_log", {
                "minimum_logical_time": 0
            }, expected_fields=["minimum_logical_time", "entries", "count"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("poll_telemetry_log")
            elif result:
                results["PASS"].append("poll_telemetry_log")
            else:
                results["FAIL"].append("poll_telemetry_log")
            
            result = await test_tool(session, "reorganize_blob", {
                "tag_id_major": 0,
                "tag_id_minor": 0,
                "blob_name": "test_blob",
                "new_score": 0.5
            }, expected_fields=["tag_id", "blob_name", "new_score", "success"])
            if result == "EXPECTED_FAILURE":
                results["EXPECTED_FAILURE"].append("reorganize_blob")
            elif result:
                results["PASS"].append("reorganize_blob")
            else:
                results["FAIL"].append("reorganize_blob")
    
    # Print summary
    print("\n" + "=" * 80)
    print("Test Summary")
    print("=" * 80)
    
    total_tests = len(results["PASS"]) + len(results["FAIL"]) + len(results["EXPECTED_FAILURE"])
    
    print(f"\n{Colors.PASS}✓ Passed: {len(results['PASS'])}{Colors.RESET}")
    for test in results["PASS"]:
        print(f"    - {test}")
    
    if results["EXPECTED_FAILURE"]:
        print(f"\n{Colors.WARN}⚠ Expected Failures: {len(results['EXPECTED_FAILURE'])}{Colors.RESET}")
        print("    (These may fail without proper runtime initialization)")
        for test in results["EXPECTED_FAILURE"]:
            print(f"    - {test}")
    
    if results["FAIL"]:
        print(f"\n{Colors.FAIL}✗ Failed: {len(results['FAIL'])}{Colors.RESET}")
        for test in results["FAIL"]:
            print(f"    - {test}")
    
    print(f"\n📊 Total Tests: {total_tests}")
    print(f"   Passed: {len(results['PASS'])}")
    print(f"   Expected Failures: {len(results['EXPECTED_FAILURE'])}")
    print(f"   Failed: {len(results['FAIL'])}")
    
    # Exit with appropriate code
    if results["FAIL"]:
        sys.exit(1)
    else:
        sys.exit(0)

if __name__ == "__main__":
    asyncio.run(main())




##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
"""
Implements a viewer frame for compass_model.Array.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import wx
import wx.grid
from wx.lib.newevent import NewCommandEvent

import os
import logging

log = logging.getLogger(__name__)

from ..frame import NodeFrame
from .plot import LinePlotFrame, ContourPlotFrame


# Indicates that the slicing selection may have changed.
# These events are emitted by the SlicerPanel.
ArraySlicedEvent, EVT_ARRAY_SLICED = NewCommandEvent()

# Menu and button IDs
ID_VIS_MENU_PLOT = wx.NewId()


class GeoArrayFrame(NodeFrame):
    """
    Top-level frame displaying objects of type compass_model.Array.

    From top to bottom, has:

    1. Toolbar (see ArrayFrame.init_toolbar)
    2. SlicerPanel, with controls for changing what's displayed.
    3. An ArrayGrid, which displays the data in a spreadsheet-like view.
    """

    def __init__(self, node, pos=None):
        """ Create a new array viewer to display the node. """
        NodeFrame.__init__(self, node, size=(800, 400), title=node.display_name, pos=pos)

        self.node = node

        # Update the menu
        vis_menu = wx.Menu()
        if self.node.is_plottable():
            vis_menu.Append(ID_VIS_MENU_PLOT, "Map Data\tCtrl-D")
            self.add_menu(vis_menu, "Visualize")
        # Initialize the toolbar
        self.init_toolbar()

        # The Slicer is the panel with indexing controls
        self.slicer = SlicerPanel(self, node.shape, node.dtype.fields is not None)
        # Create the grid array
        self.grid = ArrayGrid(self, node, self.slicer)
        # Sizer for slicer and grid
        gridsizer = wx.BoxSizer(wx.VERTICAL)
        gridsizer.Add(self.slicer, 0, wx.EXPAND)
        gridsizer.Add(self.grid, 1, wx.EXPAND)
        self.view = gridsizer

        self.Bind(EVT_ARRAY_SLICED, self.on_sliced)
        if self.node.is_plottable():
            self.Bind(wx.EVT_MENU, self.on_plot, id=ID_VIS_MENU_PLOT)

        # Workaround for wxPython bug (see SlicerPanel.enable_spinctrls)
        ID_WORKAROUND_TIMER = wx.NewId()
        self.Bind(wx.EVT_TIMER, self.on_workaround_timer, id=ID_WORKAROUND_TIMER)
        self.timer = wx.Timer(self, ID_WORKAROUND_TIMER)
        self.timer.Start(100)

    def init_toolbar(self):
        """ Set up the toolbar at the top of the window. """
        t_size = (24, 24)
        plot_bmp = wx.Bitmap(os.path.join(self.icon_folder, "viz_plot_24.png"), wx.BITMAP_TYPE_ANY)

        self.toolbar = self.CreateToolBar(wx.TB_HORIZONTAL | wx.NO_BORDER | wx.TB_FLAT | wx.TB_TEXT)

        self.toolbar.SetToolBitmapSize(t_size)
        self.toolbar.AddStretchableSpace()
        if self.node.is_plottable():
            self.toolbar.AddLabelTool(ID_VIS_MENU_PLOT, "Map Data", plot_bmp,
                                      shortHelp="Map geographic data in a popup window",
                                      longHelp="Map the geographic array data in a popup window")
        self.toolbar.Realize()

    def on_sliced(self, evt):
        """ User has chosen to display a different part of the dataset. """
        self.grid.Refresh()

    def on_plot(self, evt):
        """ User has chosen to plot the current selection """
        cols = self.grid.GetSelectedCols()
        rows = self.grid.GetSelectedRows()

        # Scalar data can't be line-plotted.
        if len(self.node.shape) == 0:
            return

        # Columns in the view are selected
        if len(cols) != 0:

            # The data is compound
            if self.node.dtype.names is not None:
                names = [self.grid.GetColLabelValue(x) for x in cols]
                data = self.node[self.slicer.indices]  # -> 1D compound array
                data = [data[n] for n in names]
                f = LinePlotFrame(data, names)
                f.Show()

            # Plot multiple columns independently
            else:
                if len(self.node.shape) == 1:
                    data = [self.node[self.slicer.indices]]
                else:
                    data = [self.node[self.slicer.indices + (c,)] for c in cols]

                names = ["Col %d" % c for c in cols] if len(data) > 1 else None

                f = LinePlotFrame(data, names)
                f.Show()


        # Rows in view are selected
        elif len(rows) != 0:

            data = [self.node[self.slicer.indices + (slice(None, None, None), r)] for r in rows]
            names = ["Row %d" % r for r in rows] if len(data) > 1 else None

            f = LinePlotFrame(data, names)
            f.Show()


        # No row or column selection.  Plot everything
        else:

            data = self.node[self.slicer.indices]

            # The data is compound
            if self.node.dtype.names is not None:
                names = [self.grid.GetColLabelValue(x) for x in xrange(self.grid.GetNumberCols())]
                data = [data[n] for n in names]
                f = LinePlotFrame(data, names)
                f.Show()

            # Plot 1D
            elif len(self.node.shape) == 1:
                f = LinePlotFrame([data])
                f.Show()

            # Plot 2D
            else:
                f = ContourPlotFrame(data, extent=self.node.extent)
                f.Show()

    def on_workaround_timer(self, evt):
        """ See slicer.enable_spinctrls docs """
        self.timer.Destroy()
        self.slicer.enable_spinctrls()


class SlicerPanel(wx.Panel):
    """
    Holds controls for data access.

    Consult the "indices" property, which returns a tuple of indices that
    prefix the array.  This will be RANK-2 elements long, unless hasfields
    is true, in which case it will be RANK-1 elements long.
    """

    @property
    def indices(self):
        """ A tuple of integer indices appropriate for slicing.

        Will be RANK-2 elements long, RANK-1 if compound data is in use
        (hasfields == True).
        """
        return tuple([x.GetValue() for x in self.spincontrols])

    def __init__(self, parent, shape, hasfields):
        """ Create a new slicer panel.

        parent:     The wxPython parent window
        shape:      Shape of the data to visualize
        hasfields:  If True, the data is compound and the grid can only
                    display one axis.  So, we should display an extra spinbox.
        """
        wx.Panel.__init__(self, parent)

        self.shape = shape
        self.hasfields = hasfields
        self.spincontrols = []

        # Rank of the underlying array
        rank = len(shape)

        # Rank displayable in the grid.  If fields are present, they occupy
        # the columns, so the data displayed is actually 1-D.
        visible_rank = 1 if hasfields else 2

        sizer = wx.BoxSizer(wx.HORIZONTAL)  # Will arrange the SpinCtrls

        if rank > visible_rank:
            infotext = wx.StaticText(self, wx.ID_ANY, "Array Indexing: ")
            sizer.Add(infotext, 0, flag=wx.EXPAND | wx.ALL, border=10)

            for idx in xrange(rank - visible_rank):
                sc = wx.SpinCtrl(self, max=shape[idx] - 1, value="0", min=0)
                sizer.Add(sc, 0, flag=wx.EXPAND | wx.ALL, border=10)
                sc.Disable()
                self.spincontrols.append(sc)

        self.SetSizer(sizer)

        self.Bind(wx.EVT_SPINCTRL, self.on_spin)

    def enable_spinctrls(self):
        """ Unlock the spin controls.

        Because of a bug in wxPython on Mac, by default the first spin control
        has bizarre contents (and control focus) when the panel starts up.
        Call this after a short delay (e.g. 100 ms) to enable indexing.
        """
        for sc in self.spincontrols:
            sc.Enable()

    def on_spin(self, evt):
        """ Spinbox value changed; notify parent to refresh the grid. """
        wx.PostEvent(self, ArraySlicedEvent(self.GetId()))


class ArrayGrid(wx.grid.Grid):
    """
    Grid class to display the Array.

    Cell contents and appearance are handled by the table model in ArrayTable.
    """

    def __init__(self, parent, node, slicer):
        wx.grid.Grid.__init__(self, parent)
        table = ArrayTable(node, slicer)
        self.SetTable(table, True)

        # Column selection is always allowed
        selmode = wx.grid.Grid.wxGridSelectColumns

        # Row selection is forbidden for compound types, and for
        # scalar/1-D datasets
        if node.dtype.names is None and len(node.shape) > 1:
            selmode |= wx.grid.Grid.wxGridSelectRows

        self.SetSelectionMode(selmode)


class LRUTileCache(object):
    """
        Simple tile-based LRU cache which goes between the Grid and
        the Array object.  Caches tiles along the last 1 or 2 dimensions
        of a dataset.

        Access is via __getitem__.  Because this class exists specifically
        to support point-based callbacks for the Grid, arguments may
        only be indices, not slices.
    """

    TILESIZE = 100  # Tiles will have shape (100,) or (100, 100)
    MAXTILES = 50  # Max number of tiles to retain in the cache

    def __init__(self, arr):
        """ *arr* is anything implementing compass_model.Array """
        import collections
        self.cache = collections.OrderedDict()
        self.arr = arr

    def __getitem__(self, args):
        """ Restricted to an index or tuple of indices. """

        if not isinstance(args, tuple):
            args = (args,)

        # Split off the last 1 or 2 dimensions
        coarse_position, fine_position = args[0:-2], args[-2:]

        def clip(x):
            """ Round down to nearest TILESIZE; takes e.g. 181 -> 100 """
            return (x // self.TILESIZE) * self.TILESIZE

        # Tuple with index of tile corner
        tile_key = coarse_position + tuple(clip(x) for x in fine_position)

        # Slice which will be applied to dataset to retrieve tile
        tile_slice = coarse_position + tuple(slice(clip(x), clip(x) + self.TILESIZE) for x in fine_position)

        # Index applied to tile to retrieve the desired data point
        tile_data_index = tuple(x % self.TILESIZE for x in fine_position)

        # Case 1: Add tile to cache, ejecting oldest tile if needed
        if not tile_key in self.cache:

            if len(self.cache) >= self.MAXTILES:
                self.cache.popitem(last=False)

            tile = self.arr[tile_slice]
            self.cache[tile_key] = tile

        # Case 2: Mark the tile as recently accessed
        else:
            tile = self.cache.pop(tile_key)
            self.cache[tile_key] = tile

        return tile[tile_data_index]


class ArrayTable(wx.grid.PyGridTableBase):
    """
    "Table" class which provides data and metadata for the grid to display.

    The methods defined here define the contents of the table, as well as
    the number of rows, columns and their values.
    """

    def __init__(self, node, slicer):
        """ Create a new Table instance for use with a grid control.

        node:     An compass_model.Array implementation instance.
        slicer:   An instance of SlicerPanel, so we can see what indices the
                  user has requested.
        """
        wx.grid.PyGridTableBase.__init__(self)

        self.node = node
        self.slicer = slicer

        self.rank = len(node.shape)
        self.names = node.dtype.names

        self.cache = LRUTileCache(self.node)

    def GetNumberRows(self):
        """ Callback for number of rows displayed by the grid control """
        if self.rank == 0:
            return 1
        return self.node.shape[-1]

    def GetNumberCols(self):
        """ Callback for number of columns displayed by the grid control.

        Note that if compound data is in use, columns display the field names.
        """
        if self.names is not None:
            return len(self.names)
        if self.rank < 2:
            return 1
        return self.node.shape[-2]

    def GetValue(self, row, col):
        """ Callback which provides data to the Grid.

        row, col:   Integers giving row and column position (0-based).
        """
        # Scalar case
        if self.rank == 0:
            data = self.node[()]
            if self.names is None:
                return data
            return data[col]

        # 1D case
        if self.rank == 1:
            data = self.cache[row]
            if self.names is None:
                return data
            return data[self.names[col]]

        # ND case.  Watch out for compound mode!
        if self.names is None:
            args = self.slicer.indices + (col, row)
        else:
            args = self.slicer.indices + (row,)

        data = self.cache[args]
        if self.names is None:
            return data
        return data[self.names[col]]

    def GetRowLabelValue(self, row):
        """ Callback for row labels.

        Row number is used unless the data is scalar.
        """
        if self.rank == 0:
            return "Value"
        return str(row)

    def GetColLabelValue(self, col):
        """ Callback for column labels.

        Column number is used, except for scalar or 1D data, or if we're
        displaying field names in the columns.
        """
        if self.names is not None:
            return self.names[col]
        if self.rank == 0 or self.rank == 1:
            return "Value"
        return str(col)



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Matplotlib window with toolbar.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import wx

# matplotlib backend is already set by viewer.py to WXAgg
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigCanvas
from matplotlib.backends.backend_wx import NavigationToolbar2Wx as NavigationToolbar

import logging
log = logging.getLogger(__name__)

from ..frame import BaseFrame

ID_VIEW_CMAP_JET = wx.ID_ANY  # default
ID_VIEW_CMAP_BONE = wx.ID_ANY
ID_VIEW_CMAP_GIST_EARTH = wx.ID_ANY
ID_VIEW_CMAP_OCEAN = wx.ID_ANY
ID_VIEW_CMAP_RAINBOW = wx.ID_ANY
ID_VIEW_CMAP_RDYLGN = wx.ID_ANY
ID_VIEW_CMAP_WINTER = wx.ID_ANY
ID_SAVE_AS = wx.ID_ANY
ID_COPY = wx.ID_ANY
ID_ZOOM_IN = wx.ID_ANY
ID_ZOOM_OUT = wx.ID_ANY
ID_ZOOM_FIT = wx.ID_ANY


class PlotFrame(BaseFrame):
    """ Base class for Matplotlib plot windows.

    Override draw_figure() to plot your figure on the provided axes.
    """

    def __init__(self, data, title="a title"):
        """ Create a new Matplotlib plotting window for a 1D line plot """

        log.debug(self.__class__.__name__)
        super(PlotFrame, self).__init__(id=wx.ID_ANY, title=title, size=(800, 400))

        self.data = data

        self.panel = wx.Panel(self)

        self.dpi = 100
        self.fig = Figure((6.0, 4.0), dpi=self.dpi)
        self.canvas = FigCanvas(self.panel, -1, self.fig)

        self.axes = self.fig.add_subplot(111)
        self.axes.set_xlabel('')
        self.axes.set_ylabel('')

        self.toolbar = NavigationToolbar(self.canvas)

        self.vbox = wx.BoxSizer(wx.VERTICAL)
        self.vbox.Add(self.canvas, 1, wx.LEFT | wx.TOP | wx.GROW)
        self.vbox.Add(self.toolbar, 0, wx.EXPAND)

        self.panel.SetSizer(self.vbox)
        self.vbox.Fit(self)

        # Add standard toolbar buttons
        self.toolbar.AddTool(wx.ID_SAVE, "Save", 
                            wx.ArtProvider.GetBitmap(wx.ART_FILE_SAVE, wx.ART_TOOLBAR),
                            "Save plot to file")
        
        self.toolbar.AddTool(wx.ID_COPY, "Copy", 
                            wx.ArtProvider.GetBitmap(wx.ART_COPY, wx.ART_TOOLBAR),
                            "Copy plot to clipboard")
        
        self.toolbar.AddTool(wx.ID_ZOOM_IN, "Zoom In", 
                            wx.ArtProvider.GetBitmap(wx.ART_ADD_BOOKMARK, wx.ART_TOOLBAR),
                            "Zoom in")
        
        self.toolbar.AddTool(wx.ID_ZOOM_OUT, "Zoom Out", 
                            wx.ArtProvider.GetBitmap(wx.ART_DEL_BOOKMARK, wx.ART_TOOLBAR),
                            "Zoom out")
        
        self.toolbar.AddTool(wx.ID_ZOOM_FIT, "Fit", 
                            wx.ArtProvider.GetBitmap(wx.ART_FIND, wx.ART_TOOLBAR),
                            "Fit plot to window")
        
        self.toolbar.Realize()

        # Bind events
        self.Bind(wx.EVT_TOOL, self.on_save, id=wx.ID_SAVE)
        self.Bind(wx.EVT_TOOL, self.on_copy, id=wx.ID_COPY)
        self.Bind(wx.EVT_TOOL, self.on_zoom_in, id=wx.ID_ZOOM_IN)
        self.Bind(wx.EVT_TOOL, self.on_zoom_out, id=wx.ID_ZOOM_OUT)
        self.Bind(wx.EVT_TOOL, self.on_zoom_fit, id=wx.ID_ZOOM_FIT)

        self.draw_figure()

    def draw_figure(self):
        raise NotImplementedError


class LinePlotFrame(PlotFrame):
    def __init__(self, data, names=None, title="Line Plot"):
        self.names = names
        self.complex_mode = 'magnitude'  # Default mode for complex data
        PlotFrame.__init__(self, data, title)

        # Add complex data controls if needed
        if self._has_complex_data():
            self._add_complex_controls()

    def _has_complex_data(self):
        """Check if any of the data arrays contain complex numbers"""
        for d in self.data:
            if np.iscomplexobj(d):
                return True
        return False

    def _add_complex_controls(self):
        """Add menu for complex data visualization options"""
        complex_menu = wx.Menu()

        item_mag = complex_menu.AppendRadioItem(wx.ID_ANY, "Magnitude")
        item_phase = complex_menu.AppendRadioItem(wx.ID_ANY, "Phase")
        item_real = complex_menu.AppendRadioItem(wx.ID_ANY, "Real Part")
        item_imag = complex_menu.AppendRadioItem(wx.ID_ANY, "Imaginary Part")
        item_both = complex_menu.AppendRadioItem(wx.ID_ANY, "Real && Imaginary")

        self.add_menu(complex_menu, "Complex Display")

        # Check magnitude by default
        item_mag.Check(True)

        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('magnitude'), id=item_mag.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('phase'), id=item_phase.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('real'), id=item_real.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('imag'), id=item_imag.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('both'), id=item_both.GetId())

    def _on_complex_mode(self, mode):
        """Handle complex display mode change"""
        self.complex_mode = mode
        self.axes.clear()
        self.draw_figure()
        self.canvas.draw()

    def draw_figure(self):
        lines = []
        plot_names = []

        for idx, d in enumerate(self.data):
            name = self.names[idx] if self.names and idx < len(self.names) else f"Data {idx}"

            if np.iscomplexobj(d):
                # Handle complex data based on mode
                if self.complex_mode == 'magnitude':
                    lines.append(self.axes.plot(np.abs(d), label=f"{name} (magnitude)")[0])
                    plot_names.append(f"{name} (magnitude)")
                elif self.complex_mode == 'phase':
                    lines.append(self.axes.plot(np.angle(d), label=f"{name} (phase)")[0])
                    plot_names.append(f"{name} (phase)")
                    self.axes.set_ylabel('Phase (radians)')
                elif self.complex_mode == 'real':
                    lines.append(self.axes.plot(d.real, label=f"{name} (real)")[0])
                    plot_names.append(f"{name} (real)")
                elif self.complex_mode == 'imag':
                    lines.append(self.axes.plot(d.imag, label=f"{name} (imag)")[0])
                    plot_names.append(f"{name} (imag)")
                elif self.complex_mode == 'both':
                    lines.append(self.axes.plot(d.real, label=f"{name} (real)", linestyle='-')[0])
                    lines.append(self.axes.plot(d.imag, label=f"{name} (imag)", linestyle='--')[0])
                    plot_names.append(f"{name} (real)")
                    plot_names.append(f"{name} (imag)")
            else:
                # Regular real data
                lines.append(self.axes.plot(d, label=name)[0])
                plot_names.append(name)

        if plot_names:
            self.axes.legend()

class LineXYPlotFrame(PlotFrame):
    def __init__(self, data, names=None, title="Line XY Plot"):
        self.names = names
        self.complex_mode = 'magnitude'  # Default mode for complex data
        PlotFrame.__init__(self, data, title)

        # Add complex data controls if needed
        if self._has_complex_data():
            self._add_complex_controls()

    def _has_complex_data(self):
        """Check if any of the data arrays contain complex numbers"""
        for d in self.data:
            if np.iscomplexobj(d):
                return True
        return False

    def _add_complex_controls(self):
        """Add menu for complex data visualization options"""
        complex_menu = wx.Menu()

        item_mag = complex_menu.AppendRadioItem(wx.ID_ANY, "Magnitude")
        item_phase = complex_menu.AppendRadioItem(wx.ID_ANY, "Phase")
        item_real = complex_menu.AppendRadioItem(wx.ID_ANY, "Real Part")
        item_imag = complex_menu.AppendRadioItem(wx.ID_ANY, "Imaginary Part")

        self.add_menu(complex_menu, "Complex Display")

        # Check magnitude by default
        item_mag.Check(True)

        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('magnitude'), id=item_mag.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('phase'), id=item_phase.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('real'), id=item_real.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('imag'), id=item_imag.GetId())

    def _on_complex_mode(self, mode):
        """Handle complex display mode change"""
        self.complex_mode = mode
        self.axes.clear()
        self.draw_figure()
        self.canvas.draw()

    def _process_complex_data(self, d):
        """Convert complex data to real based on current mode"""
        if not np.iscomplexobj(d):
            return d

        if self.complex_mode == 'magnitude':
            return np.abs(d)
        elif self.complex_mode == 'phase':
            return np.angle(d)
        elif self.complex_mode == 'real':
            return d.real
        elif self.complex_mode == 'imag':
            return d.imag
        return np.abs(d)  # Default to magnitude

    def draw_figure(self):
        # Process X data (first column)
        x_data = self._process_complex_data(self.data[0])

        self.axes.set_xlabel(self.names[0] if self.names else "X")
        if len(self.data)==2:
            # a simple X-Y plot using 2 columns
            ylabel = self.names[1] if self.names and len(self.names) > 1 else "Y"
            if np.iscomplexobj(self.data[1]):
                ylabel += f" ({self.complex_mode})"
            self.axes.set_ylabel(ylabel)

        lines = []
        for idx, d in enumerate(self.data[1::], start=1):
            y_data = self._process_complex_data(d)
            label = self.names[idx] if self.names and idx < len(self.names) else f"Y{idx}"
            if np.iscomplexobj(d):
                label += f" ({self.complex_mode})"
            lines.append(self.axes.plot(x_data, y_data, label=label)[0])

        if self.names is not None and len(lines) > 0:
            self.axes.legend()

class ContourPlotFrame(PlotFrame):
    def __init__(self, data, names=None, title="Contour Plot"):
        # need to be set before calling the parent (need for plotting)
        self.colormap = "jet"
        self.cb = None  # matplotlib color-bar
        self.complex_mode = 'magnitude'  # Default mode for complex data

        PlotFrame.__init__(self, data, title)

        # Add complex data controls if needed
        if np.iscomplexobj(self.data):
            self._add_complex_controls()

        self.cmap_menu = wx.Menu()
        self.cmap_menu.Append(ID_VIEW_CMAP_JET, "Jet", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_BONE, "Bone", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_GIST_EARTH, "Gist Earth", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_OCEAN, "Ocean", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_RAINBOW, "Rainbow", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_RDYLGN, "Red-Yellow-Green", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_WINTER, "Winter", kind=wx.ITEM_RADIO)
        self.add_menu(self.cmap_menu, "Colormap")

        self.Bind(wx.EVT_MENU, self.on_cmap_jet, id=ID_VIEW_CMAP_JET)
        self.Bind(wx.EVT_MENU, self.on_cmap_bone, id=ID_VIEW_CMAP_BONE)
        self.Bind(wx.EVT_MENU, self.on_cmap_gist_earth, id=ID_VIEW_CMAP_GIST_EARTH)
        self.Bind(wx.EVT_MENU, self.on_cmap_ocean, id=ID_VIEW_CMAP_OCEAN)
        self.Bind(wx.EVT_MENU, self.on_cmap_rainbow, id=ID_VIEW_CMAP_RAINBOW)
        self.Bind(wx.EVT_MENU, self.on_cmap_rdylgn, id=ID_VIEW_CMAP_RDYLGN)
        self.Bind(wx.EVT_MENU, self.on_cmap_winter, id=ID_VIEW_CMAP_WINTER)

        self.status_bar = wx.StatusBar(self, -1)
        self.status_bar.SetFieldsCount(2)
        self.SetStatusBar(self.status_bar)

        self.canvas.mpl_connect('motion_notify_event', self.update_status_bar)
        self.canvas.Bind(wx.EVT_ENTER_WINDOW, self.change_cursor)

    def on_cmap_jet(self, evt):
        log.debug("cmap: jet")
        self.colormap = "jet"
        self._refresh_plot()

    def on_cmap_bone(self, evt):
        log.debug("cmap: bone")
        self.colormap = "bone"
        self._refresh_plot()

    def on_cmap_gist_earth(self, evt):
        log.debug("cmap: gist_earth")
        self.colormap = "gist_earth"
        self._refresh_plot()

    def on_cmap_ocean(self, evt):
        log.debug("cmap: ocean")
        self.colormap = "ocean"
        self._refresh_plot()

    def on_cmap_rainbow(self, evt):
        log.debug("cmap: rainbow")
        self.colormap = "rainbow"
        self._refresh_plot()

    def on_cmap_rdylgn(self, evt):
        log.debug("cmap: RdYlGn")
        self.colormap = "RdYlGn"
        self._refresh_plot()

    def on_cmap_winter(self, evt):
        log.debug("cmap: winter")
        self.colormap = "winter"
        self._refresh_plot()

    def _add_complex_controls(self):
        """Add menu for complex data visualization options"""
        complex_menu = wx.Menu()

        item_mag = complex_menu.AppendRadioItem(wx.ID_ANY, "Magnitude")
        item_phase = complex_menu.AppendRadioItem(wx.ID_ANY, "Phase")
        item_real = complex_menu.AppendRadioItem(wx.ID_ANY, "Real Part")
        item_imag = complex_menu.AppendRadioItem(wx.ID_ANY, "Imaginary Part")

        self.add_menu(complex_menu, "Complex Display")

        # Check magnitude by default
        item_mag.Check(True)

        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('magnitude'), id=item_mag.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('phase'), id=item_phase.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('real'), id=item_real.GetId())
        self.Bind(wx.EVT_MENU, lambda e: self._on_complex_mode('imag'), id=item_imag.GetId())

    def _on_complex_mode(self, mode):
        """Handle complex display mode change"""
        self.complex_mode = mode
        self.axes.clear()
        self._refresh_plot()

    def _refresh_plot(self):
        self.draw_figure()
        self.canvas.draw()

    def _process_complex_data(self, data):
        """Convert complex data to real based on current mode"""
        if not np.iscomplexobj(data):
            return data

        if self.complex_mode == 'magnitude':
            return np.abs(data)
        elif self.complex_mode == 'phase':
            return np.angle(data)
        elif self.complex_mode == 'real':
            return data.real
        elif self.complex_mode == 'imag':
            return data.imag
        return np.abs(data)  # Default to magnitude

    def draw_figure(self):
        max_elements = 500  # don't attempt plot more than 500x500 elements
        rows = self.data.shape[0]
        cols = self.data.shape[1]
        row_stride = rows // max_elements + 1
        col_stride = cols // max_elements + 1

        # Process complex data if needed
        plot_data = self._process_complex_data(self.data[::row_stride, ::col_stride])

        xx = np.arange(0, self.data.shape[1], col_stride)
        yy = np.arange(0, self.data.shape[0], row_stride)
        img = self.axes.contourf(xx, yy, plot_data, 25, cmap=plt.cm.get_cmap(self.colormap))
        self.axes.set_aspect('equal')
        if self.cb:
            self.cb.on_mappable_changed(img)
        else:
            self.cb = plt.colorbar(img, ax=self.axes)
        self.cb.ax.tick_params(labelsize=8)

        # Update colorbar label if complex data
        if np.iscomplexobj(self.data):
            self.cb.set_label(self.complex_mode.capitalize())

    def change_cursor(self, event):
        self.canvas.SetCursor(wx.StockCursor(wx.CURSOR_CROSS))

    def update_status_bar(self, event):
        msg = str()
        if event.inaxes:
            x, y = int(event.xdata), int(event.ydata)
            if 0 <= y < self.data.shape[0] and 0 <= x < self.data.shape[1]:
                z = self.data[y, x]
                if np.iscomplexobj(self.data):
                    # Show complex value information
                    mag = np.abs(z)
                    phase = np.angle(z)
                    msg = f"x={x}, y={y}, z={z:.3f}, |z|={mag:.3f}, ∠z={phase:.3f} rad"
                else:
                    msg = "x= %d, y= %d, z= %f" % (x, y, z)
        self.status_bar.SetStatusText(msg, 1)



import test_stl_ext as t
import typing
import pytest
import sys
from common import collect, skip_on_pypy

# Some helper declaration to check types across different Python versions
if sys.version_info < (3, 9):
    TYPING_DICT = "typing.Dict"
    TYPING_SET = "typing.Set"
    TYPING_ABC = "typing"
    TYPING_ABSTRACT_SET = "typing.AbstractSet"
else:
    TYPING_DICT = "dict"
    TYPING_SET = "set"
    TYPING_ABC = "collections.abc"
    TYPING_ABSTRACT_SET = "collections.abc.Set"


def optional(arg: str, /) -> str:
    if sys.version_info < (3, 10):
        return "typing.Optional[" + arg + "]"
    else:
        return arg + " | " + "None"


def union(*args: str) -> str:
    if sys.version_info < (3, 10):
        return "typing.Union[" + ", ".join(args) + "]"
    else:
        return " | ".join(args)


@pytest.fixture
def clean():
    collect()
    t.reset()


def assert_stats(**kwargs):
    collect()
    for k, v in t.stats().items():
        fail = False
        if k in kwargs:
            if v != kwargs[k]:
                fail = True
        elif v != 0:
            fail = True
        if fail:
            raise Exception(f"Mismatch for key {k}: {t.stats()}")


# ------------------------------------------------------------------
# The following aren't strictly STL tests, but they are helpful in
# ensuring that move constructors/copy constructors of bound C++ types
# are properly triggered, which the STL type casters depend on.
# ------------------------------------------------------------------


def test01_movable_return(clean):
    assert t.return_movable().value == 5
    assert_stats(default_constructed=1, move_constructed=1, destructed=2)


def test02_movable_return_ptr(clean):
    assert t.return_movable_ptr().value == 5
    assert_stats(default_constructed=1, destructed=1)


def test03_movable_in_value(clean):
    s = t.Movable()
    t.movable_in_value(s)
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, copy_constructed=1, destructed=2)


def test04_movable_in_lvalue_ref(clean):
    s = t.Movable()
    t.movable_in_lvalue_ref(s)
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, destructed=1)


def test05_movable_in_ptr(clean):
    s = t.Movable()
    t.movable_in_ptr(s)
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, destructed=1)


def test06_movable_in_rvalue_ref(clean):
    s = t.Movable()
    t.movable_in_rvalue_ref(s)
    assert s.value == 0
    del s
    assert_stats(default_constructed=1, move_constructed=1, destructed=2)


def test07_copyable_return(clean):
    assert t.return_copyable().value == 5
    assert_stats(default_constructed=1, copy_constructed=1, destructed=2)


def test08_copyable_return_ptr(clean):
    assert t.return_copyable_ptr().value == 5
    assert_stats(default_constructed=1, destructed=1)


def test09_copyable_in_value(clean):
    s = t.Copyable()
    t.copyable_in_value(s)
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, copy_constructed=1, destructed=2)


def test10_copyable_in_lvalue_ref(clean):
    s = t.Copyable()
    t.copyable_in_lvalue_ref(s)
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, destructed=1)


def test11_copyable_in_ptr(clean):
    s = t.Copyable()
    t.copyable_in_ptr(s)
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, destructed=1)


def test12_copyable_in_rvalue_ref(clean):
    s = t.Copyable()
    t.copyable_in_rvalue_ref(s)
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, copy_constructed=1, destructed=2)


# ------------------------------------------------------------------


def test13_tuple_movable_return(clean):
    assert t.tuple_return_movable()[0].value == 5
    assert_stats(default_constructed=1, move_constructed=2, destructed=3)


def test14_tuple_movable_return_ptr(clean):
    assert t.return_movable_ptr().value == 5
    assert_stats(default_constructed=1, destructed=1)


def test15_tuple_movable_in_value(clean):
    s = t.Movable()
    t.tuple_movable_in_value((s,))
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, copy_constructed=1, destructed=2)


def test16_tuple_movable_in_lvalue_ref(clean):
    s = t.Movable()
    t.tuple_movable_in_lvalue_ref((s,))
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, destructed=1)


def test17_tuple_movable_in_lvalue_ref_2(clean):
    s = t.Movable()
    t.tuple_movable_in_lvalue_ref_2((s,))
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, copy_constructed=1, destructed=2)


def test18_tuple_movable_in_ptr(clean):
    s = t.Movable()
    t.tuple_movable_in_ptr((s,))
    assert s.value == 5
    del s
    assert_stats(default_constructed=1, destructed=1)


def test19_tuple_movable_in_rvalue_ref(clean):
    s = t.Movable()
    t.tuple_movable_in_rvalue_ref((s,))
    assert s.value == 0
    del s
    assert_stats(default_constructed=1, move_constructed=1, destructed=2)


def test20_tuple_movable_in_rvalue_ref_2(clean):
    s = t.Movable()
    t.tuple_movable_in_rvalue_ref_2((s,))
    assert s.value == 5
    del s
    assert_stats(
        default_constructed=1, copy_constructed=1, move_constructed=1, destructed=3
    )


# ------------------------------------------------------------------


def test21_tuple_pair_basic():
    assert t.empty_tuple(()) == ()
    assert t.swap_tuple((1, 2.5)) == (2.5, 1)
    assert t.swap_pair((1, 2.5)) == (2.5, 1)


# ------------------------------------------------------------------


def test22_vec_return_movable(clean):
    for i, x in enumerate(t.vec_return_movable()):
        assert x.value == i
    del x
    assert_stats(value_constructed=10, move_constructed=10, destructed=20)


def test23_vec_return_copyable(clean):
    for i, x in enumerate(t.vec_return_copyable()):
        assert x.value == i
    del x
    assert_stats(value_constructed=10, copy_constructed=20, destructed=30)


def test24_vec_movable_in_value(clean):
    t.vec_movable_in_value([t.Movable(i) for i in range(10)])
    assert_stats(value_constructed=10, copy_constructed=10, destructed=20)

    # Test that None values don't cause a crash
    with pytest.raises(TypeError):
        t.vec_movable_in_value([None])


def test25_vec_movable_in_value(clean):
    t.vec_copyable_in_value([t.Copyable(i) for i in range(10)])
    assert_stats(value_constructed=10, copy_constructed=10, destructed=20)


def test26_vec_movable_in_lvalue_ref(clean):
    t.vec_movable_in_lvalue_ref([t.Movable(i) for i in range(10)])
    assert_stats(value_constructed=10, copy_constructed=10, destructed=20)


def test27_vec_movable_in_ptr_2(clean):
    t.vec_movable_in_ptr_2([t.Movable(i) for i in range(10)])
    assert_stats(value_constructed=10, destructed=10)

    # Test that None values are permitted when casting to pointer;
    # instead we reach 'if (x.size() != 10) fail();' in the bound function
    with pytest.raises(RuntimeError):
        t.vec_movable_in_ptr_2([None])


def test28_vec_movable_in_rvalue_ref(clean):
    t.vec_movable_in_rvalue_ref([t.Movable(i) for i in range(10)])
    assert_stats(value_constructed=10, copy_constructed=10, destructed=20)


def test29_opaque_vector():
    f = t.float_vec()
    assert f.size() == 0
    assert isinstance(f, t.float_vec)
    f.push_back(1)
    assert f.size() == 1


def test30_std_function():
    assert t.return_empty_function() is None
    assert t.return_function()(3) == 8
    assert t.call_function(lambda x: 5 + x, 3) == 8

    with pytest.raises(TypeError) as excinfo:
        assert t.call_function(5, 3) == 8
    assert "incompatible function arguments" in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        assert t.call_function(lambda x, y: x + y, 3) == 8
    assert "missing 1 required positional argument" in str(excinfo.value)

    l = []

    def f():
        l.append(1)

    f2 = t.return_void_function(f)
    f2()
    assert l == [1]


def test31_std_function_roundtrip():
    def f():
        l.append(1)

    f2 = t.return_void_function(f)
    assert f2 is f


# cpyext reference cycles are not supported, see https://foss.heptapod.net/pypy/pypy/-/issues/3849
@skip_on_pypy
def test32_std_function_gc():
    # Test class -> function -> class cyclic reference
    class A(t.FuncWrapper):
        pass

    assert t.FuncWrapper.alive == 0
    a = A()
    assert t.FuncWrapper.alive == 1
    del a
    collect()
    assert t.FuncWrapper.alive == 0

    # t.FuncWrapper is a C extension type with a custom property 'f', which
    # can store Python function objects It implements the tp_traverse
    # callback so that reference cycles arising from this function object
    # can be detected.

    class Test(t.FuncWrapper):
        def __init__(self):
            super().__init__()

            # The constructor creates a closure, which references 'self'
            # and assigns it to the 'self.f' member.
            # This creates a cycle self -> self.f -> self
            def f():
                print(self.f)

            self.f = f


    # The Test class declared above inherits from 'FuncWrapper'.
    # This class keeps track of how many references are alive at
    # any point to help track down leak issues.
    assert t.FuncWrapper.alive == 0
    b = Test()
    assert t.FuncWrapper.alive == 1
    del b
    collect()
    assert t.FuncWrapper.alive == 0


def test33_vec_type_check():
    with pytest.raises(TypeError) as excinfo:
        t.vec_movable_in_value(0)


def test34_list():
    assert t.identity_list([]) == []
    assert t.identity_list([1, 2, 3]) == [1, 2, 3]
    assert t.identity_list(()) == []
    assert t.identity_list((1, 2, 3)) == [1, 2, 3]


def test35_string_and_string_view():
    assert t.identity_string("") == ""
    assert t.identity_string("orange") == "orange"
    assert t.identity_string("橘子") == "橘子"
    assert t.identity_string("ส้ม") == "ส้ม"
    assert t.identity_string("البرتقالي") == "البرتقالي"
    assert t.identity_string("🍊") == "🍊"

    assert t.identity_string_view("") == ""
    assert t.identity_string_view("orange") == "orange"
    assert t.identity_string_view("橘子") == "橘子"
    assert t.identity_string_view("ส้ม") == "ส้ม"
    assert t.identity_string_view("البرتقالي") == "البرتقالي"
    assert t.identity_string_view("🍊") == "🍊"


def test36_std_optional_copyable(clean):
    t.optional_copyable(t.Copyable())
    opt_copyable = optional("test_stl_ext.Copyable")
    assert t.optional_copyable.__doc__ == (
        f"optional_copyable(x: {opt_copyable}) -> None"
    )
    assert_stats(default_constructed=1, copy_constructed=1, destructed=2)


def test37_std_optional_copyable_ptr(clean):
    t.optional_copyable_ptr(t.Copyable())
    opt_copyable = optional("test_stl_ext.Copyable")
    assert t.optional_copyable_ptr.__doc__ == (
        f"optional_copyable_ptr(x: {opt_copyable}) -> None"
    )
    assert_stats(default_constructed=1, destructed=1)


def test38_std_optional_none():
    t.optional_none(None)
    assert t.optional_cstr(None) == "none"
    assert t.optional_cstr("hi") == "hi"


def test39_std_optional_ret_opt_movable(clean):
    assert t.optional_ret_opt_movable().value == 5
    opt_movable = optional("test_stl_ext.Movable")
    assert t.optional_ret_opt_movable.__doc__ == (
        f"optional_ret_opt_movable() -> {opt_movable}"
    )
    assert_stats(default_constructed=1, move_constructed=2, destructed=3)


def test40_std_optional_ret_opt_movable_ptr(clean):
    assert t.optional_ret_opt_movable_ptr().value == 5
    assert_stats(default_constructed=1, destructed=1)


def test41_std_optional_ret_opt_none():
    assert t.optional_ret_opt_none() is None


def test42_std_optional_unbound_type():
    for method_name in (
        "optional_unbound_type",
        "optional_unbound_type_with_nullopt_as_default",
    ):
        method = getattr(t, method_name)
        assert method(3) == 3
        assert method(None) is None
        assert method() is None
        opt_int = optional("int")
        assert method.__doc__ == (f"{method_name}(x: {opt_int} = None) -> {opt_int}")


def test42a_std_optional_non_assignable():
    assert t.optional_non_assignable(t.NonAssignable()).value == 5


def test43_std_variant_copyable(clean):
    t.variant_copyable(t.Copyable())
    t.variant_copyable(5)
    arg_t = union('test_stl_ext.Copyable', 'int')
    assert t.variant_copyable.__doc__ == (
        f"variant_copyable(arg: {arg_t}, /) -> None"
    )
    assert_stats(default_constructed=3, copy_assigned=1, destructed=3)


def test44_std_variant_copyable_none(clean):
    t.variant_copyable_none(t.Copyable())
    t.variant_copyable_none(5)
    t.variant_copyable_none(None)
    arg_t = optional(union("int", "test_stl_ext.Copyable"))
    assert t.variant_copyable_none.__doc__ == (
        f"variant_copyable_none(x: {arg_t}) -> None"
    )
    assert_stats(default_constructed=1, copy_constructed=1, destructed=2)


def test45_std_variant_copyable_ptr(clean):
    t.variant_copyable_ptr(t.Copyable())
    t.variant_copyable_ptr(5)
    arg_t = union("test_stl_ext.Copyable", "int")
    assert t.variant_copyable_ptr.__doc__ == (
        f"variant_copyable_ptr(arg: {arg_t}, /) -> None"
    )
    assert_stats(default_constructed=1, destructed=1)


def test46_std_variant_copyable_ptr_none(clean):
    t.variant_copyable_ptr_none(t.Copyable())
    t.variant_copyable_ptr_none(5)
    t.variant_copyable_ptr_none(None)
    arg_t = optional(union("test_stl_ext.Copyable", "int"))
    assert t.variant_copyable_ptr_none.__doc__ == (
        f"variant_copyable_ptr_none(x: {arg_t}) -> None"
    )
    assert_stats(default_constructed=1, destructed=1)


def test47_std_variant_ret_var_copyable():
    assert t.variant_ret_var_copyable().value == 5
    rv_t = union("test_stl_ext.Copyable", "int")
    assert t.variant_ret_var_copyable.__doc__ == (
        f"variant_ret_var_copyable() -> {rv_t}"
    )


def test48_std_variant_ret_var_none():
    assert t.variant_ret_var_none() is None

    rv_t = union("None", "test_stl_ext.Copyable", "int")
    assert t.variant_ret_var_none.__doc__ == (f"variant_ret_var_none() -> {rv_t}")


def test49_std_variant_unbound_type():
    assert t.variant_unbound_type() is None
    assert t.variant_unbound_type(None) is None
    assert t.variant_unbound_type([5]) == [5]
    assert t.variant_unbound_type((1, 2, 3)) == (1, 2, 3)
    assert t.variant_unbound_type(5) == 5
    arg_t = optional(union("list", "tuple", "int"))
    rv_t = union("None", "list", "tuple", "int")
    assert t.variant_unbound_type.__doc__ == (
        f"variant_unbound_type(x: {arg_t} = None) -> {rv_t}"
    )

def test49b_std_variant_nondefault(clean):
    assert t.variant_nondefault(t.NonDefaultConstructible(10)) == 10
    assert t.variant_nondefault(20) == -20
    assert_stats(
        value_constructed=1, # constructed in NonDefaultConstructible pyobject
        copy_constructed=1, # copied into type_caster variant member
        move_constructed=1, # moved into function argument value
        destructed=3,
    )

def test50_map_return_movable_value():
    for i, (k, v) in enumerate(sorted(t.map_return_movable_value().items())):
        assert k == chr(ord("a") + i)
        assert v.value == i
        assert t.map_return_movable_value.__doc__ == (
            f"map_return_movable_value() -> {TYPING_DICT}[str, test_stl_ext.Movable]"
        )


def test51_map_return_copyable_value():
    for i, (k, v) in enumerate(sorted(t.map_return_copyable_value().items())):
        assert k == chr(ord("a") + i)
        assert v.value == i
        assert t.map_return_copyable_value.__doc__ == (
            f"map_return_copyable_value() -> {TYPING_DICT}[str, test_stl_ext.Copyable]"
        )


def test52_map_movable_in_value():
    t.map_movable_in_value(dict([(chr(ord("a") + i), t.Movable(i)) for i in range(10)]))
    assert t.map_movable_in_value.__doc__ == (
        f"map_movable_in_value(x: {TYPING_ABC}.Mapping[str, test_stl_ext.Movable]) -> None"
    )


def test53_map_copyable_in_value():
    t.map_copyable_in_value(
        dict([(chr(ord("a") + i), t.Copyable(i)) for i in range(10)])
    )
    assert t.map_copyable_in_value.__doc__ == (
        f"map_copyable_in_value(x: {TYPING_ABC}.Mapping[str, test_stl_ext.Copyable]) -> None"
    )


def test54_map_movable_in_lvalue_ref():
    t.map_movable_in_lvalue_ref(
        dict([(chr(ord("a") + i), t.Movable(i)) for i in range(10)])
    )
    assert t.map_movable_in_lvalue_ref.__doc__ == (
        f"map_movable_in_lvalue_ref(x: {TYPING_ABC}.Mapping[str, test_stl_ext.Movable]) -> None"
    )


def test55_map_movable_in_rvalue_ref():
    t.map_movable_in_rvalue_ref(
        dict([(chr(ord("a") + i), t.Movable(i)) for i in range(10)])
    )
    assert t.map_movable_in_rvalue_ref.__doc__ == (
        f"map_movable_in_rvalue_ref(x: {TYPING_ABC}.Mapping[str, test_stl_ext.Movable]) -> None"
    )


def test56_map_movable_in_ptr():
    t.map_movable_in_ptr(dict([(chr(ord("a") + i), t.Movable(i)) for i in range(10)]))
    assert t.map_movable_in_ptr.__doc__ == (
        f"map_movable_in_ptr(x: {TYPING_ABC}.Mapping[str, test_stl_ext.Movable]) -> None"
    )


def test57_map_return_readonly_value():
    for i, (k, v) in enumerate(sorted(t.map_return_readonly_value().map.items())):
        assert k == chr(ord("a") + i)
        assert v == i
    assert t.map_return_readonly_value.__doc__ == (
        "map_return_readonly_value() -> test_stl_ext.StructWithReadonlyMap"
    )


def test58_array():
    o = t.array_out()
    assert isinstance(o, list) and o == [1, 2, 3]
    assert t.array_in([1, 2, 3]) == 6
    assert t.array_in((1, 2, 3)) == 6
    with pytest.raises(TypeError) as excinfo:
        assert t.array_in((1, 2, 3, 4)) == 6
    assert "incompatible function arguments" in str(excinfo.value)


def test59_map_movable_in_failure():
    with pytest.raises(TypeError) as excinfo:
        t.map_copyable_in_value({1: 2})
    assert "incompatible function arguments" in str(excinfo.value)


def test60_set_return_value():
    for i, k in enumerate(sorted(t.set_return_value())):
        assert k == chr(ord("a") + i)
    for i, k in enumerate(sorted(t.unordered_set_return_value())):
        assert k == chr(ord("a") + i)
    assert t.set_return_value.__doc__ == (f"set_return_value() -> {TYPING_SET}[str]")
    assert t.unordered_set_return_value.__doc__ == (
        f"unordered_set_return_value() -> {TYPING_SET}[str]"
    )


def test61_set_in_value():
    t.set_in_value(set([chr(ord("a") + i) for i in range(10)]))
    t.unordered_set_in_value(set([chr(ord("a") + i) for i in range(10)]))
    assert t.set_in_value.__doc__ == (
        f"set_in_value(x: {TYPING_ABSTRACT_SET}[str]) -> None"
    )
    assert t.unordered_set_in_value.__doc__ == (
        f"unordered_set_in_value(x: {TYPING_ABSTRACT_SET}[str]) -> None"
    )


def test62_set_in_lvalue_ref():
    t.set_in_lvalue_ref(set([chr(ord("a") + i) for i in range(10)]))
    assert t.set_in_lvalue_ref.__doc__ == (
        f"set_in_lvalue_ref(x: {TYPING_ABSTRACT_SET}[str]) -> None"
    )


def test63_set_in_rvalue_ref():
    t.set_in_rvalue_ref(set([chr(ord("a") + i) for i in range(10)]))
    assert t.set_in_rvalue_ref.__doc__ == (
        f"set_in_rvalue_ref(x: {TYPING_ABSTRACT_SET}[str]) -> None"
    )


def test64_set_in_failure():
    with pytest.raises(TypeError) as excinfo:
        t.set_in_value(set([i for i in range(10)]))
    assert "incompatible function arguments" in str(excinfo.value)


def test65_class_with_movable_field(clean):
    cwmf = t.ClassWithMovableField()
    m1 = t.Movable(1)
    m2 = t.Movable(2)

    assert_stats(value_constructed=2)

    cwmf.movable = [m1, m2]

    assert_stats(value_constructed=2, copy_constructed=2)

    del m1, m2
    collect()

    assert_stats(value_constructed=2, copy_constructed=2, destructed=2)

    del cwmf
    collect()

    assert_stats(value_constructed=2, copy_constructed=2, destructed=4)


def test66_replace_extension():
    from pathlib import Path

    if not hasattr(t, "replace_extension"):
        pytest.skip("filesystem tests not compiled")

    filename = Path("test.txt")
    assert t.replace_extension(filename, ".obj") == filename.with_suffix(".obj")

    filename = Path("🍊.html")
    assert t.replace_extension(filename, ".svg") == filename.with_suffix(".svg")

    class PseudoStrPath:
        def __fspath__(self):
            return "foo/bar"

    class PseudoBytesPath:
        def __fspath__(self):
            return b"foo/bar"

    assert t.parent_path(Path("foo/bar")) == Path("foo")
    assert t.parent_path("foo/bar") == Path("foo")
    assert t.parent_path(b"foo/bar") == Path("foo")
    assert t.parent_path(PseudoStrPath()) == Path("foo")
    assert t.parent_path(PseudoBytesPath()) == Path("foo")


def test67_vector_bool():
    bool_vector = [True, False, True, False]
    result = t.flip_vector_bool(bool_vector)
    assert result == [not x for x in bool_vector]


def test68_complex_value():
    # double: 64bits
    assert t.complex_value_double(1.0) == complex(1.0)
    assert t.complex_value_double(1.0j) == 1.0j
    assert t.complex_value_double(0.0) == complex(0.0)
    assert t.complex_value_double(0.0j) == 0.0j
    assert t.complex_value_double(0) == complex(0.0)
    assert t.complex_value_double_nc(1.0 + 2.0j) == 1.0 + 2.0j
    assert t.complex_value_double_nc(1.0j) == 1.0j
    assert t.complex_value_double_nc(0.0 + 0.0j) == complex(0.0)
    assert t.complex_value_double_nc(0.0j) == 0.0j
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_double_nc(0)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_double_nc(0.0)

    # float: 32bits
    assert t.complex_value_float(1.0) == complex(1.0)
    assert t.complex_value_float(1.0j) == 1.0j
    assert t.complex_value_float(0.0) == complex(0.0)
    assert t.complex_value_float(0.0j) == 0.0j
    assert t.complex_value_float(0) == complex(0.0)
    assert t.complex_value_float_nc(1.0 + 2.0j) == 1.0 + 2.0j
    assert t.complex_value_float_nc(1.0j) == 1.0j
    assert t.complex_value_float_nc(0.0 + 0.0j) == complex(0.0)
    assert t.complex_value_float_nc(0.0j) == 0.0j
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_float_nc(0)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_float_nc(0.0)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_float_nc(1.1 + 2.0j)  # Inexact narrowing conversion
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_float_nc(1.0 + 2.1j)  # Inexact narrowing conversion

    val_64 = 2.7 - 3.2j
    val_32 = 2.700000047683716 - 3.200000047683716j
    assert val_64 != val_32

    assert t.complex_value_double(val_32) == val_32
    assert t.complex_value_double(val_64) == val_64
    assert t.complex_value_double_nc(val_32) == val_32
    assert t.complex_value_double_nc(val_64) == val_64

    assert t.complex_value_float(val_32) == val_32
    assert t.complex_value_float(val_64) == val_32
    assert t.complex_value_float_nc(val_32) == val_32
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_float_nc(val_64)  # Inexact narrowing conversion

    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_float([])

    class MyInt(int):
        def __new__(cls, value):
            return super().__new__(cls, value)

    assert t.complex_value_double(MyInt(7)) == complex(7.0)
    assert t.complex_value_float(MyInt(7)) == complex(7.0)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_double_nc(MyInt(7))
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_float_nc(MyInt(7))

    class MyComplex:
        def __init__(self, real, imag):
            self.re = real
            self.im = imag
        def __complex__(self):
            return complex(self.re, self.im)

    assert t.complex_value_double(MyComplex(1, 2)) == complex(1.0, 2.0)
    assert t.complex_value_float(MyComplex(1, 2)) == complex(1.0, 2.0)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_double_nc(MyComplex(1, 2))
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_float_nc(MyComplex(1, 2))

    class MyComplexSubclass(complex): ...

    assert t.complex_value_double(MyComplexSubclass(1, 2)) == complex(1.0, 2.0)
    assert t.complex_value_float(MyComplexSubclass(1, 2)) == complex(1.0, 2.0)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_double_nc(MyComplexSubclass(1, 2))
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_value_float_nc(MyComplexSubclass(1, 2))

    try:
        import numpy as np

        assert t.complex_value_float(np.complex64(val_32)) == val_32
        assert t.complex_value_float(np.complex64(val_64)) == val_32
        assert t.complex_value_double(np.complex64(val_32)) == val_32
        assert t.complex_value_double(np.complex64(val_64)) == val_32

        assert t.complex_value_float(np.complex128(val_32)) == val_32
        assert t.complex_value_float(np.complex128(val_64)) == val_32
        assert t.complex_value_double(np.complex128(val_32)) == val_32
        assert t.complex_value_double(np.complex128(val_64)) == val_64

        with pytest.raises(TypeError, match="incompatible function arguments"):
            t.complex_value_double_nc(np.complex128(val_64))
        with pytest.raises(TypeError, match="incompatible function arguments"):
            t.complex_value_float_nc(np.complex64(val_32))

    except ImportError:
        pass


def test69_complex_array():
    val1_64 = 2.7 - 3.2j
    val1_32 = 2.700000047683716 - 3.200000047683716j
    val2_64 = 3.1415
    val2_32 = 3.1414999961853027 + 0j

    # test 64 bit casts
    assert t.complex_array_double([val1_64, -1j, val2_64]) == [
        val1_64,
        -0 - 1j,
        val2_64,
    ]

    # test 32 bit casts
    assert t.complex_array_float([val1_64, -1j, val2_64]) == [
        val1_32,
        (-0 - 1j),
        val2_32,
    ]

    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.complex_array_float([[]])

    try:
        import numpy as np

        # test 64 bit casts
        assert t.complex_array_double(np.array([val1_64, -1j, val2_64])) == [
            val1_64,
            -0 - 1j,
            val2_64,
        ]
        assert t.complex_array_double(
            np.array([val1_64, -1j, val2_64], dtype=np.complex128)
        ) == [val1_64, -0 - 1j, val2_64]
        assert t.complex_array_double(
            np.array([val1_64, -1j, val2_64], dtype=np.complex64)
        ) == [val1_32, -0 - 1j, val2_32]

        # test 32 bit casts
        assert t.complex_array_float(np.array([val1_64, -1j, val2_64])) == [
            val1_32,
            (-0 - 1j),
            val2_32,
        ]
        assert t.complex_array_float(
            np.array([val1_64, -1j, val2_64], dtype=np.complex128)
        ) == [val1_32, (-0 - 1j), val2_32]
        assert t.complex_array_float(
            np.array([val1_64, -1j, val2_64], dtype=np.complex64)
        ) == [val1_32, (-0 - 1j), val2_32]
    except ImportError:
        pass


def test70_vec_char():
    assert isinstance(t.vector_str("123"), str)
    assert isinstance(t.vector_str(["123", "345"]), list)
    assert t.vector_optional_str(["abc", None]) == ["abc", None]


def test71_null_input():
    with pytest.raises(TypeError):
        t.vec_movable_in_value([None])
    with pytest.raises(TypeError):
        t.map_copyable_in_value({"a": None})

@skip_on_pypy # PyPy fails this test on Windows :-(
def test72_wstr():
    assert t.pass_wstr('🎈') == '🎈'

def test73_bad_input_to_set():
    with pytest.raises(TypeError):
        t.set_in_value(None)

def test74_variant_implicit_conversions():
    event = t.IDHavingEvent()
    assert event.id is None
    event.id = t.BasicID1(78)
    assert type(event.id) is t.BasicID1
